"""skylet: a daemon running on the head node of a cluster."""

import concurrent.futures
import os
import time

import grpc

import sky
from sky import sky_logging
from sky.schemas.generated import autostopv1_pb2_grpc
from sky.skylet import constants
from sky.skylet import events
from sky.skylet import services

# Use the explicit logger name so that the logger is under the
# `sky.skylet.skylet` namespace when executed directly, so as
# to inherit the setup from the `sky` logger.
logger = sky_logging.init_logger('sky.skylet.skylet')
logger.info(f'Skylet started with version {constants.SKYLET_VERSION}; '
            f'SkyPilot v{sky.__version__} (commit: {sky.__commit__})')

EVENTS = [
    events.AutostopEvent(),
    events.JobSchedulerEvent(),
    # The managed job update event should be after the job update event.
    # Otherwise, the abnormal managed job status update will be delayed
    # until the next job update event.
    events.ManagedJobEvent(),
    # This is for monitoring controller job status. If it becomes
    # unhealthy, this event will correctly update the controller
    # status to CONTROLLER_FAILED.
    events.ServiceUpdateEvent(pool=False),
    # Status refresh for pool.
    events.ServiceUpdateEvent(pool=True),
    # Report usage heartbeat every 10 minutes.
    events.UsageHeartbeatReportEvent(),
]


def start_grpc_server(port: int = constants.SKYLET_GRPC_PORT) -> grpc.Server:
    """Start the gRPC server."""
    # This is the default value in Python 3.8 - 3.12,
    # putting it here for visibility.
    # TODO(kevin): Determine the optimal max number of threads.
    max_workers = min(32, (os.cpu_count() or 1) + 4)
    server = grpc.server(
        concurrent.futures.ThreadPoolExecutor(max_workers=max_workers))

    autostopv1_pb2_grpc.add_AutostopServiceServicer_to_server(
        services.AutostopServiceImpl(), server)

    listen_addr = f'127.0.0.1:{port}'
    server.add_insecure_port(listen_addr)

    server.start()
    logger.info(f'gRPC server started on {listen_addr}')

    return server


def run_event_loop():
    """Run the existing event loop."""

    while True:
        time.sleep(events.EVENT_CHECKING_INTERVAL_SECONDS)
        for event in EVENTS:
            event.run()


def main():
    grpc_server = start_grpc_server()
    try:
        run_event_loop()
    except KeyboardInterrupt:
        logger.info('Shutting down skylet...')
    finally:
        grpc_server.stop(grace=5)


if __name__ == '__main__':
    main()



"""The SkyPilot package."""
import os
import subprocess
from typing import Optional
import urllib.request

# Replaced with the current commit when building the wheels.
_SKYPILOT_COMMIT_SHA = '523104e8e33153f9aba88897ea9a472513bbd9da'


def _get_git_commit():
    if 'SKYPILOT_COMMIT_SHA' not in _SKYPILOT_COMMIT_SHA:
        # This is a release build, so we don't need to get the commit hash from
        # git, as it's already been set.
        return _SKYPILOT_COMMIT_SHA

    # This is a development build (pip install -e .), so we need to get the
    # commit hash from git.
    try:
        cwd = os.path.dirname(__file__)
        commit_hash = subprocess.check_output(
            ['git', 'rev-parse', 'HEAD'],
            cwd=cwd,
            universal_newlines=True,
            stderr=subprocess.DEVNULL).strip()
        changes = subprocess.check_output(['git', 'status', '--porcelain'],
                                          cwd=cwd,
                                          universal_newlines=True,
                                          stderr=subprocess.DEVNULL).strip()
        if changes:
            commit_hash += '-dirty'
        return commit_hash
    except Exception:  # pylint: disable=broad-except
        return _SKYPILOT_COMMIT_SHA


__commit__ = _get_git_commit()
__version__ = '1.0.0.dev20250819'
__root_dir__ = os.path.dirname(os.path.abspath(__file__))


# ---------------------- Proxy Configuration ---------------------- #
def _set_http_proxy_env_vars() -> None:
    urllib_proxies = dict(urllib.request.getproxies())

    def set_proxy_env_var(proxy_var: str, urllib_var: Optional[str]):
        """Sets proxy env vars in os.environ, consulting urllib if needed.

        Logic:
        - If either PROXY_VAR or proxy_var is set in os.environ, set both to the
          same value in os.environ.
        - Else, if urllib_var is set in urllib.request.getproxies(), use that
          value to set PROXY_VAR and proxy_var in os.environ.

        Although many of our underlying libraries are case-insensitive when it
        comes to proxy environment variables, some are not. This has happened to
        GCP's SDK not respecting certain VPN-related proxy env vars.

        This function ensures that both the upper and lower case versions of the
        proxy environment variables are set if either is set to ensure maximum
        compatibility.
        """
        # Check for the uppercase version first
        proxy = os.getenv(proxy_var.upper(), os.getenv(proxy_var.lower()))
        if proxy is None and urllib_var is not None:
            proxy = urllib_proxies.get(urllib_var)

        if proxy is not None:
            os.environ[proxy_var.lower()] = proxy
            os.environ[proxy_var.upper()] = proxy

    set_proxy_env_var('http_proxy', 'http')
    set_proxy_env_var('https_proxy', 'https')
    set_proxy_env_var('all_proxy', None)


_set_http_proxy_env_vars()

# ----------------------------------------------------------------- #

# Keep this order to avoid cyclic imports
# pylint: disable=wrong-import-position
from sky import backends
from sky import clouds
from sky.admin_policy import AdminPolicy
from sky.admin_policy import MutatedUserRequest
from sky.admin_policy import UserRequest
from sky.catalog import list_accelerators
from sky.client.sdk import api_cancel
from sky.client.sdk import api_info
from sky.client.sdk import api_login
from sky.client.sdk import api_server_logs
from sky.client.sdk import api_start
from sky.client.sdk import api_status
from sky.client.sdk import api_stop
from sky.client.sdk import autostop
from sky.client.sdk import cancel
from sky.client.sdk import cost_report
from sky.client.sdk import down
from sky.client.sdk import download_logs
from sky.client.sdk import endpoints
from sky.client.sdk import exec  # pylint: disable=redefined-builtin
from sky.client.sdk import get
from sky.client.sdk import job_status
from sky.client.sdk import launch
from sky.client.sdk import optimize
from sky.client.sdk import queue
from sky.client.sdk import reload_config
from sky.client.sdk import start
from sky.client.sdk import status
from sky.client.sdk import stop
from sky.client.sdk import storage_delete
from sky.client.sdk import storage_ls
from sky.client.sdk import stream_and_get
from sky.client.sdk import tail_logs
from sky.dag import Dag
from sky.data import Storage
from sky.data import StorageMode
from sky.data import StoreType
from sky.jobs import ManagedJobStatus
from sky.optimizer import Optimizer
from sky.resources import Resources
from sky.skylet.job_lib import JobStatus
from sky.task import Task
from sky.utils.common import OptimizeTarget
from sky.utils.common import StatusRefreshMode
from sky.utils.config_utils import Config
from sky.utils.registry import CLOUD_REGISTRY
from sky.utils.registry import JOBS_RECOVERY_STRATEGY_REGISTRY
from sky.utils.status_lib import ClusterStatus

# Aliases.
IBM = clouds.IBM
AWS = clouds.AWS
Azure = clouds.Azure
Cudo = clouds.Cudo
GCP = clouds.GCP
Lambda = clouds.Lambda
SCP = clouds.SCP
Kubernetes = clouds.Kubernetes
K8s = Kubernetes
OCI = clouds.OCI
Paperspace = clouds.Paperspace
RunPod = clouds.RunPod
Vast = clouds.Vast
Vsphere = clouds.Vsphere
Fluidstack = clouds.Fluidstack
Nebius = clouds.Nebius
Hyperbolic = clouds.Hyperbolic

__all__ = [
    '__version__',
    'AWS',
    'Azure',
    'Cudo',
    'GCP',
    'IBM',
    'Kubernetes',
    'K8s',
    'Lambda',
    'OCI',
    'Paperspace',
    'RunPod',
    'Vast',
    'SCP',
    'Vsphere',
    'Fluidstack',
    'Nebius',
    'Hyperbolic',
    'Optimizer',
    'OptimizeTarget',
    'backends',
    'list_accelerators',
    '__root_dir__',
    'Storage',
    'StorageMode',
    'StoreType',
    'ClusterStatus',
    'JobStatus',
    'ManagedJobStatus',
    'StatusRefreshMode',
    # APIs
    'Dag',
    'Task',
    'Resources',
    # core APIs
    'optimize',
    'launch',
    'exec',
    'reload_config',
    # core APIs
    'status',
    'start',
    'stop',
    'down',
    'autostop',
    'cost_report',
    'endpoints',
    # core APIs Job Management
    'queue',
    'cancel',
    'tail_logs',
    'download_logs',
    'job_status',
    # core APIs Storage Management
    'storage_ls',
    'storage_delete',
    # API server APIs
    'get',
    'stream_and_get',
    'api_status',
    'api_cancel',
    'api_info',
    'api_login',
    'api_start',
    'api_stop',
    'api_server_logs',
    # Admin Policy
    'UserRequest',
    'MutatedUserRequest',
    'AdminPolicy',
    'Config',
    # Registry
    'CLOUD_REGISTRY',
    'JOBS_RECOVERY_STRATEGY_REGISTRY',
]



"""Cudo Compute instance provisioning."""

import time
from typing import Any, Dict, List, Optional, Tuple

from sky import sky_logging
from sky.provision import common
from sky.provision.cudo import cudo_machine_type
from sky.provision.cudo import cudo_wrapper
from sky.utils import status_lib

POLL_INTERVAL = 10

logger = sky_logging.init_logger(__name__)


def _filter_instances(cluster_name_on_cloud: str,
                      status_filters: Optional[List[str]]) -> Dict[str, Any]:
    instances = cudo_wrapper.list_instances()
    possible_names = [
        f'{cluster_name_on_cloud}-head', f'{cluster_name_on_cloud}-worker'
    ]

    filtered_nodes = {}
    for instance_id, instance in instances.items():
        if (status_filters is not None and
                instance['status'] not in status_filters):
            continue
        if instance.get('name') in possible_names:
            filtered_nodes[instance_id] = instance
    return filtered_nodes


def _get_head_instance_id(instances: Dict[str, Any]) -> Optional[str]:
    head_instance_id = None
    for inst_id, inst in instances.items():
        if inst['name'].endswith('-head'):
            head_instance_id = inst_id
            break
    return head_instance_id


def run_instances(region: str, cluster_name_on_cloud: str,
                  config: common.ProvisionConfig) -> common.ProvisionRecord:
    """Runs instances for the given cluster."""

    pending_status = ['pend', 'init', 'prol', 'boot']

    while True:
        instances = _filter_instances(cluster_name_on_cloud, pending_status)
        if not instances:
            break
        logger.info(f'Waiting for {len(instances)} instances to be ready.')
        time.sleep(POLL_INTERVAL)
    exist_instances = _filter_instances(cluster_name_on_cloud, ['runn'])
    head_instance_id = _get_head_instance_id(exist_instances)

    to_start_count = config.count - len(exist_instances)
    if to_start_count < 0:
        raise RuntimeError(
            f'Cluster {cluster_name_on_cloud} already has '
            f'{len(exist_instances)} nodes, but {config.count} are required.')
    if to_start_count == 0:
        if head_instance_id is None:
            raise RuntimeError(
                f'Cluster {cluster_name_on_cloud} has no head node.')
        logger.info(f'Cluster {cluster_name_on_cloud} already has '
                    f'{len(exist_instances)} nodes, no need to start more.')
        return common.ProvisionRecord(provider_name='cudo',
                                      cluster_name=cluster_name_on_cloud,
                                      region=region,
                                      zone=None,
                                      head_instance_id=head_instance_id,
                                      resumed_instance_ids=[],
                                      created_instance_ids=[])

    created_instance_ids = []
    public_key = config.node_config['AuthorizedKey']
    instance_type = config.node_config['InstanceType']
    spec = cudo_machine_type.get_spec_from_instance(instance_type, region)
    gpu_count = int(float(spec['gpu_count']))
    vcpu_count = int(spec['vcpu_count'])
    memory_gib = int(spec['mem_gb'])
    gpu_model = spec['gpu_model']
    try:
        cudo_wrapper.vm_available(to_start_count, gpu_count, gpu_model, region,
                                  memory_gib, vcpu_count)
    except Exception as e:
        logger.warning(f'run_instances: {e}')
        raise
    for _ in range(to_start_count):

        node_type = 'head' if head_instance_id is None else 'worker'
        try:
            instance_id = cudo_wrapper.launch(
                name=f'{cluster_name_on_cloud}-{node_type}',
                ssh_key=public_key,
                data_center_id=region,
                machine_type=spec['machine_type'],
                memory_gib=memory_gib,
                vcpu_count=vcpu_count,
                gpu_count=gpu_count,
                tags={},
                disk_size=config.node_config['DiskSize'])
        except Exception as e:  # pylint: disable=broad-except
            logger.warning(f'run_instances error: {e}')
            raise
        logger.info(f'Launched instance {instance_id}.')
        created_instance_ids.append(instance_id)
        if head_instance_id is None:
            head_instance_id = instance_id

    # Wait for instances to be ready.
    retries = 60  # times 10 second
    results = {}
    for instance_id in created_instance_ids:
        for _ in range(retries):
            logger.info('Waiting for instance(s) to be ready '
                        f'{instance_id}')
            vm = cudo_wrapper.get_instance(instance_id)
            if vm['vm']['short_state'] == 'runn':
                results[id] = vm['vm']
                break
            else:
                time.sleep(POLL_INTERVAL)

    assert head_instance_id is not None, 'head_instance_id should not be None'
    return common.ProvisionRecord(provider_name='cudo',
                                  cluster_name=cluster_name_on_cloud,
                                  region=region,
                                  zone=None,
                                  head_instance_id=head_instance_id,
                                  resumed_instance_ids=[],
                                  created_instance_ids=created_instance_ids)


def wait_instances(region: str, cluster_name_on_cloud: str,
                   state: Optional[status_lib.ClusterStatus]) -> None:
    # Waiting is done in run_instances
    del region, cluster_name_on_cloud, state


def stop_instances(
    cluster_name_on_cloud: str,
    provider_config: Optional[Dict[str, Any]] = None,
    worker_only: bool = False,
) -> None:
    raise NotImplementedError()


def terminate_instances(
    cluster_name_on_cloud: str,
    provider_config: Optional[Dict[str, Any]] = None,
    worker_only: bool = False,
) -> None:
    """See sky/provision/__init__.py"""
    del provider_config
    instances = _filter_instances(cluster_name_on_cloud, None)
    for inst_id, inst in instances.items():
        if worker_only and inst['name'].endswith('-head'):
            continue
        logger.debug(f'Terminating Cudo instance {inst_id}.'
                     f'{inst}')
        cudo_wrapper.remove(inst_id)


def get_cluster_info(
        region: str,
        cluster_name_on_cloud: str,
        provider_config: Optional[Dict[str, Any]] = None) -> common.ClusterInfo:
    del region
    nodes = _filter_instances(cluster_name_on_cloud, ['runn', 'pend'])
    instances: Dict[str, List[common.InstanceInfo]] = {}
    head_instance_id = None
    for node_id, node_info in nodes.items():
        instances[node_id] = [
            common.InstanceInfo(
                instance_id=node_id,
                internal_ip=node_info['internal_ip'],
                external_ip=node_info['external_ip'],
                tags=node_info['tags'],
            )
        ]
        if node_info['name'].endswith('-head'):
            head_instance_id = node_id

    return common.ClusterInfo(instances=instances,
                              head_instance_id=head_instance_id,
                              provider_name='cudo',
                              provider_config=provider_config)


def query_instances(
    cluster_name: str,
    cluster_name_on_cloud: str,
    provider_config: Optional[Dict[str, Any]] = None,
    non_terminated_only: bool = True,
) -> Dict[str, Tuple[Optional['status_lib.ClusterStatus'], Optional[str]]]:
    """See sky/provision/__init__.py"""
    del cluster_name  # unused
    assert provider_config is not None, (cluster_name_on_cloud, provider_config)
    instances = _filter_instances(cluster_name_on_cloud, None)

    status_map = {
        'init': status_lib.ClusterStatus.INIT,
        'pend': status_lib.ClusterStatus.INIT,
        'prol': status_lib.ClusterStatus.INIT,
        'boot': status_lib.ClusterStatus.INIT,
        'runn': status_lib.ClusterStatus.UP,
        'stop': status_lib.ClusterStatus.STOPPED,
        'susp': status_lib.ClusterStatus.STOPPED,
        'done': status_lib.ClusterStatus.STOPPED,
        'poff': status_lib.ClusterStatus.STOPPED,
    }
    statuses: Dict[str, Tuple[Optional['status_lib.ClusterStatus'],
                              Optional[str]]] = {}
    for inst_id, inst in instances.items():
        status = status_map[inst['status']]
        if non_terminated_only and status is None:
            continue
        statuses[inst_id] = (status, None)
    return statuses


def open_ports(
    cluster_name_on_cloud: str,
    ports: List[str],
    provider_config: Optional[Dict[str, Any]] = None,
) -> None:
    del cluster_name_on_cloud, ports, provider_config
    # Cudo has all ports open by default. Nothing to do here.
    return


def cleanup_ports(
    cluster_name_on_cloud: str,
    ports: List[str],
    provider_config: Optional[Dict[str, Any]] = None,
) -> None:
    del cluster_name_on_cloud, ports, provider_config



"""Managed jobs."""
import pathlib

from sky.jobs.client.sdk import cancel
from sky.jobs.client.sdk import dashboard
from sky.jobs.client.sdk import download_logs
from sky.jobs.client.sdk import launch
from sky.jobs.client.sdk import pool_apply
from sky.jobs.client.sdk import pool_down
from sky.jobs.client.sdk import pool_status
from sky.jobs.client.sdk import pool_sync_down_logs
from sky.jobs.client.sdk import pool_tail_logs
from sky.jobs.client.sdk import queue
from sky.jobs.client.sdk import tail_logs
from sky.jobs.constants import JOBS_CLUSTER_NAME_PREFIX_LENGTH
from sky.jobs.constants import JOBS_CONTROLLER_LOGS_DIR
from sky.jobs.constants import JOBS_CONTROLLER_TEMPLATE
from sky.jobs.constants import JOBS_CONTROLLER_YAML_PREFIX
from sky.jobs.constants import JOBS_TASK_YAML_PREFIX
from sky.jobs.recovery_strategy import StrategyExecutor
from sky.jobs.state import ManagedJobStatus
from sky.jobs.utils import dump_managed_job_queue
from sky.jobs.utils import format_job_table
from sky.jobs.utils import load_managed_job_queue
from sky.jobs.utils import ManagedJobCodeGen

pathlib.Path(JOBS_TASK_YAML_PREFIX).expanduser().parent.mkdir(parents=True,
                                                              exist_ok=True)
__all__ = [
    # Constants
    'JOBS_CONTROLLER_TEMPLATE',
    'JOBS_CONTROLLER_YAML_PREFIX',
    'JOBS_TASK_YAML_PREFIX',
    'JOBS_CONTROLLER_LOGS_DIR',
    # Enums
    'ManagedJobStatus',
    # Core
    'cancel',
    'launch',
    'queue',
    'tail_logs',
    'dashboard',
    'download_logs',
    # utils
    'ManagedJobCodeGen',
    'format_job_table',
    'dump_managed_job_queue',
    'load_managed_job_queue',
    'StrategyExecutor',
]



"""SkyServeController: the central controller of SkyServe.

Responsible for autoscaling and replica management.
"""
import contextlib
import logging
import os
import threading
import time
import traceback
from typing import Any, Dict, List

import colorama
import fastapi
from fastapi import responses
import uvicorn

from sky import serve
from sky import sky_logging
from sky.serve import autoscalers
from sky.serve import replica_managers
from sky.serve import serve_state
from sky.serve import serve_utils
from sky.utils import common_utils
from sky.utils import ux_utils

logger = sky_logging.init_logger(__name__)


class AutoscalerInfoFilter(logging.Filter):

    def filter(self, record: logging.LogRecord) -> bool:
        message = record.getMessage()
        return not ('GET' in message and '200' in message and
                    '/autoscaler/info' in message)


class SkyServeController:
    """SkyServeController: control everything about replica.

    This class is responsible for:
        - Starting and terminating the replica monitor and autoscaler.
        - Providing the HTTP Server API for SkyServe to communicate with.
    """

    def __init__(self, service_name: str, service_spec: serve.SkyServiceSpec,
                 service_task_yaml: str, host: str, port: int) -> None:
        self._service_name = service_name
        self._replica_manager: replica_managers.ReplicaManager = (
            replica_managers.SkyPilotReplicaManager(
                service_name=service_name,
                spec=service_spec,
                service_task_yaml_path=service_task_yaml))
        self._autoscaler: autoscalers.Autoscaler = (
            autoscalers.Autoscaler.from_spec(service_name, service_spec))
        self._host = host
        self._port = port
        self._app = fastapi.FastAPI(lifespan=self.lifespan)

    @contextlib.asynccontextmanager
    async def lifespan(self, _: fastapi.FastAPI):
        uvicorn_access_logger = logging.getLogger('uvicorn.access')
        for handler in uvicorn_access_logger.handlers:
            handler.setFormatter(sky_logging.FORMATTER)
            handler.addFilter(AutoscalerInfoFilter())
        yield

    def _run_autoscaler(self):
        logger.info('Starting autoscaler.')
        while True:
            try:
                replica_infos = serve_state.get_replica_infos(
                    self._service_name)
                # Use the active versions set by replica manager to make
                # sure we only scale down the outdated replicas that are
                # not used by the load balancer.
                record = serve_state.get_service_from_name(self._service_name)
                assert record is not None, ('No service record found for '
                                            f'{self._service_name}')
                active_versions = record['active_versions']
                logger.info(f'All replica info: {replica_infos}')
                scaling_options = self._autoscaler.generate_scaling_decisions(
                    replica_infos, active_versions)
                for scaling_option in scaling_options:
                    logger.info(f'Scaling option received: {scaling_option}')
                    if (scaling_option.operator ==
                            autoscalers.AutoscalerDecisionOperator.SCALE_UP):
                        assert (scaling_option.target is None or isinstance(
                            scaling_option.target, dict)), scaling_option
                        self._replica_manager.scale_up(scaling_option.target)
                    else:
                        assert isinstance(scaling_option.target,
                                          int), scaling_option
                        self._replica_manager.scale_down(scaling_option.target)
            except Exception as e:  # pylint: disable=broad-except
                # No matter what error happens, we should keep the
                # monitor running.
                logger.error('Error in autoscaler: '
                             f'{common_utils.format_exception(e)}')
                with ux_utils.enable_traceback():
                    logger.error(f'  Traceback: {traceback.format_exc()}')
            time.sleep(self._autoscaler.get_decision_interval())

    def run(self) -> None:

        @self._app.get('/autoscaler/info')
        async def get_autoscaler_info() -> fastapi.Response:
            return responses.JSONResponse(content=self._autoscaler.info(),
                                          status_code=200)

        @self._app.post('/controller/load_balancer_sync')
        async def load_balancer_sync(
                request: fastapi.Request) -> fastapi.Response:
            request_data = await request.json()
            # TODO(MaoZiming): Check aggregator type.
            request_aggregator: Dict[str, Any] = request_data.get(
                'request_aggregator', {})
            timestamps: List[int] = request_aggregator.get('timestamps', [])
            logger.info(f'Received {len(timestamps)} inflight requests.')
            self._autoscaler.collect_request_information(request_aggregator)
            return responses.JSONResponse(content={
                'ready_replica_urls':
                    self._replica_manager.get_active_replica_urls()
            },
                                          status_code=200)

        @self._app.post('/controller/update_service')
        async def update_service(request: fastapi.Request) -> fastapi.Response:
            request_data = await request.json()
            try:
                version = request_data.get('version', None)
                if version is None:
                    return responses.JSONResponse(
                        content={'message': 'Error: version is not specified.'},
                        status_code=400)
                update_mode_str = request_data.get(
                    'mode', serve_utils.DEFAULT_UPDATE_MODE.value)
                update_mode = serve_utils.UpdateMode(update_mode_str)
                logger.info(f'Update to new version {version} with '
                            f'update_mode {update_mode}.')
                # The yaml with the name latest_task_yaml will be synced
                # See sky/serve/core.py::update
                latest_task_yaml = serve_utils.generate_task_yaml_file_name(
                    self._service_name, version)
                service = serve.SkyServiceSpec.from_yaml(latest_task_yaml)
                logger.info(
                    f'Update to new version version {version}: {service}')

                self._replica_manager.update_version(version,
                                                     service,
                                                     update_mode=update_mode)
                new_autoscaler = autoscalers.Autoscaler.from_spec(
                    self._service_name, service)
                if not isinstance(self._autoscaler, type(new_autoscaler)):
                    logger.info('Autoscaler type changed to '
                                f'{type(new_autoscaler)}, updating autoscaler.')
                    old_autoscaler = self._autoscaler
                    self._autoscaler = new_autoscaler
                    self._autoscaler.load_dynamic_states(
                        old_autoscaler.dump_dynamic_states())
                self._autoscaler.update_version(version,
                                                service,
                                                update_mode=update_mode)
                return responses.JSONResponse(content={'message': 'Success'},
                                              status_code=200)
            except Exception as e:  # pylint: disable=broad-except
                exception_str = common_utils.format_exception(e)
                logger.error(f'Error in update_service: {exception_str}')
                return responses.JSONResponse(content={
                    'message': 'Error',
                    'exception': exception_str,
                    'traceback': traceback.format_exc()
                },
                                              status_code=500)

        @self._app.post('/controller/terminate_replica')
        async def terminate_replica(
                request: fastapi.Request) -> fastapi.Response:
            request_data = await request.json()
            replica_id = request_data['replica_id']
            assert isinstance(replica_id,
                              int), 'Error: replica ID must be an integer.'
            purge = request_data['purge']
            assert isinstance(purge, bool), 'Error: purge must be a boolean.'
            replica_info = serve_state.get_replica_info_from_id(
                self._service_name, replica_id)
            assert replica_info is not None, (f'Error: replica '
                                              f'{replica_id} does not exist.')
            replica_status = replica_info.status

            if replica_status == serve_state.ReplicaStatus.SHUTTING_DOWN:
                return responses.JSONResponse(
                    status_code=409,
                    content={
                        'message':
                            f'Replica {replica_id} of service '
                            f'{self._service_name!r} is already in the process '
                            f'of terminating. Skip terminating now.'
                    })

            if (replica_status in serve_state.ReplicaStatus.failed_statuses()
                    and not purge):
                return responses.JSONResponse(
                    status_code=409,
                    content={
                        'message': f'{colorama.Fore.YELLOW}Replica '
                                   f'{replica_id} of service '
                                   f'{self._service_name!r} is in failed '
                                   f'status ({replica_info.status}). '
                                   f'Skipping its termination as it could '
                                   f'lead to a resource leak. '
                                   f'(Use `sky serve down '
                                   f'{self._service_name!r} --replica-id '
                                   f'{replica_id} --purge` to '
                                   'forcefully terminate the replica.)'
                                   f'{colorama.Style.RESET_ALL}'
                    })

            self._replica_manager.scale_down(replica_id, purge=purge)

            action = 'terminated' if not purge else 'purged'
            message = (f'{colorama.Fore.GREEN}Replica {replica_id} of service '
                       f'{self._service_name!r} is scheduled to be '
                       f'{action}.{colorama.Style.RESET_ALL}\n'
                       f'Please use {ux_utils.BOLD}sky serve status '
                       f'{self._service_name}{ux_utils.RESET_BOLD} '
                       f'to check the latest status.')
            return responses.JSONResponse(status_code=200,
                                          content={'message': message})

        @self._app.exception_handler(Exception)
        async def validation_exception_handler(
                request: fastapi.Request, exc: Exception) -> fastapi.Response:
            with ux_utils.enable_traceback():
                logger.error(f'Error in controller: {exc!r}')
            return responses.JSONResponse(
                status_code=500,
                content={
                    'message':
                        (f'Failed method {request.method} at URL {request.url}.'
                         f' Exception message is {exc!r}.')
                },
            )

        threading.Thread(target=self._run_autoscaler).start()

        logger.info('SkyServe Controller started on '
                    f'http://{self._host}:{self._port}. PID: {os.getpid()}')

        uvicorn.run(self._app, host=self._host, port=self._port)


# TODO(tian): Probably we should support service that will stop the VM in
# specific time period.
def run_controller(service_name: str, service_spec: serve.SkyServiceSpec,
                   service_task_yaml: str, controller_host: str,
                   controller_port: int):
    controller = SkyServeController(service_name, service_spec,
                                    service_task_yaml, controller_host,
                                    controller_port)
    controller.run()



"""Instrumentation for the API server."""

import os
import time

import fastapi
from prometheus_client import generate_latest
from prometheus_client import multiprocess
import prometheus_client as prom
import starlette.middleware.base
import uvicorn

from sky import sky_logging

logger = sky_logging.init_logger(__name__)

# Total number of API server requests, grouped by path, method, and status.
sky_apiserver_requests_total = prom.Counter(
    'sky_apiserver_requests_total',
    'Total number of API server requests',
    ['path', 'method', 'status'],
)

# Time spent processing API server requests, grouped by path, method, and
# status.
sky_apiserver_request_duration_seconds = prom.Histogram(
    'sky_apiserver_request_duration_seconds',
    'Time spent processing API server requests',
    ['path', 'method', 'status'],
    buckets=(0.01, 0.025, 0.05, 0.1, 0.25, 0.5, 1.0, 2.5, 5.0, 10.0,
             float('inf')),
)

metrics_app = fastapi.FastAPI()


@metrics_app.get('/metrics')
async def metrics() -> fastapi.Response:
    """Expose aggregated Prometheus metrics from all worker processes."""
    if os.environ.get('PROMETHEUS_MULTIPROC_DIR'):
        # In multiprocess mode, we need to collect metrics from all processes.
        registry = prom.CollectorRegistry()
        multiprocess.MultiProcessCollector(registry)
        data = generate_latest(registry)
    else:
        data = generate_latest()
    return fastapi.Response(content=data,
                            media_type=prom.CONTENT_TYPE_LATEST,
                            headers={'Cache-Control': 'no-cache'})


def build_metrics_server(host: str, port: int) -> uvicorn.Server:
    metrics_config = uvicorn.Config(
        'sky.server.metrics:metrics_app',
        host=host,
        port=port,
        workers=1,
    )
    metrics_server_instance = uvicorn.Server(metrics_config)
    return metrics_server_instance


def _get_status_code_group(status_code: int) -> str:
    """Group status codes into classes (2xx, 5xx) to reduce cardinality."""
    return f'{status_code // 100}xx'


def _is_streaming_api(path: str) -> bool:
    """Check if the path is a streaming API."""
    path = path.rstrip('/')
    return path.endswith('/logs') or path.endswith('/api/stream')


class PrometheusMiddleware(starlette.middleware.base.BaseHTTPMiddleware):
    """Middleware to collect Prometheus metrics for HTTP requests."""

    async def dispatch(self, request: fastapi.Request, call_next):
        path = request.url.path
        logger.info(f'PROM Middleware Request: {request}, {request.url.path}')
        streaming = _is_streaming_api(path)
        if not streaming:
            # Exclude streaming APIs, the duration is not meaningful.
            # TODO(aylei): measure the duration of async execution instead.
            start_time = time.time()
        method = request.method
        status_code_group = ''

        try:
            response = await call_next(request)
            status_code_group = _get_status_code_group(response.status_code)
        except Exception:  # pylint: disable=broad-except
            status_code_group = '5xx'
            raise
        finally:
            sky_apiserver_requests_total.labels(path=path,
                                                method=method,
                                                status=status_code_group).inc()
            if not streaming:
                duration = time.time() - start_time
                sky_apiserver_request_duration_seconds.labels(
                    path=path, method=method,
                    status=status_code_group).observe(duration)

        return response



"""Utility functions for handling infrastructure specifications."""
import dataclasses
from typing import Optional

from sky.utils import common_utils
from sky.utils import ux_utils

_REGION_OR_ZONE_TRUNCATION_LENGTH = 25


@dataclasses.dataclass
class InfraInfo:
    """Infrastructure information parsed from infra string.

    When a field is None, it means the field is not specified.
    """
    cloud: Optional[str] = None
    region: Optional[str] = None
    zone: Optional[str] = None

    def __init__(self,
                 cloud: Optional[str] = None,
                 region: Optional[str] = None,
                 zone: Optional[str] = None):
        assert cloud not in ['none', 'None', 'NONE'], 'cloud must be specified'
        if not cloud or cloud == '*':
            cloud = None
        if not region or region == '*':
            region = None
        if not zone or zone == '*':
            zone = None

        self.cloud = cloud
        self.region = region
        self.zone = zone

    @staticmethod
    def from_str(infra: Optional[str]) -> 'InfraInfo':
        """Parse the infra string into cloud, region, and zone components.

        The format of the infra string is `cloud`, `cloud/region`, or
        `cloud/region/zone`. Examples: `aws`, `aws/us-east-1`,
        `aws/us-east-1/us-east-1a`. For any field, you can use `*` to indicate
        that any value is acceptable.

        If `*` is used for any field, the InfraInfo will have None for that
        field.

        Args:
            infra: A string in the format of `cloud`, `cloud/region`, or
                `cloud/region/zone`. Examples: `aws`, `aws/us-east-1`,
                `aws/us-east-1/us-east-1a`.

        Returns:
            An InfraInfo object containing cloud, region, and zone information.

        Raises:
            ValueError: If the infra string is malformed.
        """
        if infra is None or not infra.strip():
            return InfraInfo()

        infra = infra.strip().strip('/')

        # Split on / to get cloud, region, zone
        parts = [p.strip() for p in infra.strip().split('/')]

        if '' in parts:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    f'Invalid infra format: {infra}. Format should not contain '
                    'empty parts (e.g., double slashes "//").')

        if not parts or not parts[0]:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    f'Invalid infra format: {infra}. Expected format is '
                    '"cloud", "cloud/region", or "cloud/region/zone".')

        cloud_name: Optional[str] = parts[0].lower()

        # Handle Kubernetes contexts specially, as they can contain slashes
        if cloud_name in ['k8s', 'kubernetes']:
            # For Kubernetes, the entire string after "k8s/" is the
            # context name (region)
            cloud_name = 'kubernetes'  # Normalize k8s to kubernetes
            region = '/'.join(parts[1:]) if len(parts) >= 2 else None
            zone = None
        elif cloud_name == 'ssh':
            # For SSH, the entire string after "ssh/" is the
            # node pool name. We prepend 'ssh-' for the internal implementation
            # which reuses the context name.
            # TODO(romilb): This is a workaround while we use the global
            # kubeconfig to store the ssh contexts.
            region = '/'.join(parts[1:]) if len(parts) >= 2 else None
            if region:
                region = f'ssh-{region}'
            zone = None
        else:
            # For non-Kubernetes clouds, continue with regular parsing
            # but be careful to only split into max 3 parts
            region_zone_parts = parts[1:]
            region = None
            zone = None
            if region_zone_parts:
                region = region_zone_parts[0]
                if len(region_zone_parts) > 1:
                    zone = region_zone_parts[1]
                if len(region_zone_parts) > 2:
                    with ux_utils.print_exception_no_traceback():
                        raise ValueError(
                            f'Invalid infra format: {infra}. Expected format '
                            'is "cloud", "cloud/region", or '
                            '"cloud/region/zone".')

        if cloud_name == '*':
            cloud_name = None
        if region == '*':
            region = None
        if zone == '*':
            zone = None
        return InfraInfo(cloud=cloud_name, region=region, zone=zone)

    def to_str(self) -> Optional[str]:
        """Formats cloud, region, and zone into an infra string.

        Args:
            cloud: The cloud object
            region: The region name
            zone: The zone name

        Returns:
            A formatted infra string, or None if cloud is None or '*'
        """
        cloud = self.cloud
        region = self.region
        zone = self.zone

        if cloud is None:
            cloud = '*'
        if region is None:
            region = '*'
        if zone is None:
            zone = '*'

        # If the cloud is ssh, we remove the ssh- prefix from the region
        # TODO(romilb): This is a workaround while we use the global
        # kubeconfig to store the ssh contexts.
        if region and region.startswith('ssh-'):
            region = region[4:]

        # Build the parts list and filter out trailing wildcards
        parts = [cloud.lower(), region, zone]
        while parts and parts[-1] == '*':
            parts.pop()

        if not parts:
            return None

        # Join the parts with '/'
        return '/'.join(parts)

    def formatted_str(self, truncate: bool = True) -> str:
        """Formats cloud, region, and zone into an infra string.

        Args:
            truncate: Whether to truncate the region or zone

        Returns:
            A formatted infra string, or None if cloud is None or '*'
        """
        if self.cloud is None or self.cloud == '*':
            return '-'

        region_or_zone = None
        if self.zone is not None and self.zone != '*':
            region_or_zone = self.zone
        elif self.region is not None and self.region != '*':
            # If using region, we remove the ssh- prefix if it exists for SSH
            # Node Pools.
            # TODO(romilb): This is a workaround while we use the global
            # kubeconfig to store the ssh contexts.
            region_or_zone = common_utils.removeprefix(self.region, 'ssh-')

        if region_or_zone is not None and truncate:
            region_or_zone = common_utils.truncate_long_string(
                region_or_zone,
                _REGION_OR_ZONE_TRUNCATION_LENGTH,
                truncate_middle=True)

        formatted_str = f'{self.cloud}'
        if region_or_zone is not None:
            formatted_str += f' ({region_or_zone})'

        return formatted_str



import pathlib
import textwrap

import pytest

from sky.exceptions import InvalidSkyPilotConfigError
from sky.task import Task


def _create_config_file(config: str, tmp_path: pathlib.Path) -> str:
    config_path = tmp_path / 'config.yaml'
    config_path.open('w', encoding='utf-8').write(config)
    return config_path


def test_empty_fields_task(tmp_path):
    config_path = _create_config_file(
        textwrap.dedent(f"""\
            name: task
            resources:
            workdir: examples/
            file_mounts:
            setup: echo "Running setup."
            num_nodes:
            run:
              # commented out, empty run
            """), tmp_path)
    task = Task.from_yaml(config_path)

    assert task.name == 'task'
    assert list(task.resources)[0].is_empty()
    assert task.file_mounts is None
    assert task.run is None
    assert task.setup == 'echo "Running setup."'
    assert task.num_nodes == 1
    assert task.workdir == 'examples/'


def test_invalid_fields_task(tmp_path):
    config_path = _create_config_file(
        textwrap.dedent(f"""\
            name: task

            not_a_valid_field:
            """), tmp_path)
    with pytest.raises(AssertionError) as e:
        Task.from_yaml(config_path)
    assert 'Invalid task args' in e.value.args[0]


def test_empty_fields_resources(tmp_path):
    config_path = _create_config_file(
        textwrap.dedent(f"""\
            resources:
                cloud:
                region:
                accelerators: V100:1
                disk_size:
                use_spot:
                cpus: 32
            """), tmp_path)
    task = Task.from_yaml(config_path)

    resources = list(task.resources)[0]
    assert resources.cloud is None
    assert resources.region is None
    assert resources.accelerators == {'V100': 1}
    assert resources.disk_size is 256
    assert resources.use_spot is False
    assert resources.cpus == '32'


def test_gpu_resources(tmp_path):
    config_path = _create_config_file(
        textwrap.dedent("""\
            resources:
                gpus: V100:1
            """), tmp_path)
    task = Task.from_yaml(config_path)

    resources = list(task.resources)[0]
    assert resources.accelerators == {'V100': 1}


def test_both_gpus_and_accelerators_raises_config_error(tmp_path):
    """Aliased and canonical names cannot both be specified in config."""
    config_path = _create_config_file(
        textwrap.dedent("""\
            resources:
                accelerators: V100:1
                gpus: [T4:4, V100:8]
            """), tmp_path)

    with pytest.raises(InvalidSkyPilotConfigError) as e:
        Task.from_yaml(config_path)
    assert e.value.args[
        0] == "Cannot specify both gpus and accelerators in config."


def test_invalid_fields_resources(tmp_path):
    config_path = _create_config_file(
        textwrap.dedent(f"""\
            resources:
                cloud: aws
                not_a_valid_field:
            """), tmp_path)
    with pytest.raises(AssertionError) as e:
        Task.from_yaml(config_path)
    assert 'Invalid resource args' in e.value.args[0]


def test_empty_fields_storage(tmp_path):
    config_path = _create_config_file(
        textwrap.dedent(f"""\
            file_mounts:
                /mystorage:
                    name: sky-dataset
                    source:
                    store:
                    persistent:
            """), tmp_path)
    task = Task.from_yaml(config_path)

    storage = task.storage_mounts['/mystorage']
    assert storage.name == 'sky-dataset'
    assert storage.source is None
    assert not storage.stores
    assert storage.persistent


def test_invalid_fields_storage(tmp_path):
    config_path = _create_config_file(
        textwrap.dedent(f"""\
            file_mounts:
                /datasets-storage:
                    name: sky-dataset
                    not_a_valid_field:
            """), tmp_path)
    with pytest.raises(AssertionError) as e:
        Task.from_yaml(config_path)
    assert 'Invalid storage args' in e.value.args[0]


def test_invalid_envs_key(tmp_path):
    config_path = _create_config_file(
        textwrap.dedent(f"""\
            envs:
                invalid_env_$_key: value
            """), tmp_path)
    with pytest.raises(ValueError) as e:
        Task.from_yaml(config_path)
    assert 'does not match any of the regexes:' in e.value.args[0]


def test_invalid_envs_type(tmp_path):
    config_path = _create_config_file(
        textwrap.dedent(f"""\
            envs:
                - env_key1: abc
                - env_key2: abc
            """), tmp_path)
    with pytest.raises(ValueError) as e:
        Task.from_yaml(config_path)
    assert 'is not of type \'dict\'' in e.value.args[0]


def test_invalid_empty_envs(tmp_path):
    config_path = _create_config_file(
        textwrap.dedent(f"""\
            envs:
                env_key1: abc
                env_key2:
            """), tmp_path)
    with pytest.raises(ValueError) as e:
        Task.from_yaml(config_path)
    assert 'Environment variable \'env_key2\' is None.' in e.value.args[0]


def test_replace_envs_in_workdir(tmpdir, tmp_path):
    config_path = _create_config_file(
        textwrap.dedent(f"""\
            envs:
                env_key1: {tmpdir}
            workdir: $env_key1
            """), tmp_path)
    task = Task.from_yaml(config_path)
    assert task.workdir == tmpdir



"""Utils for workspaces."""
import collections
from typing import Any, Dict, List

from sky import global_user_state
from sky import sky_logging

logger = sky_logging.init_logger(__name__)


def get_workspace_users(workspace_config: Dict[str, Any]) -> List[str]:
    """Get the users that should have access to a workspace.

    workspace_config is a dict with the following keys:
    - private: bool
    - allowed_users: list of user names or IDs

    This function will automatically resolve the user names to IDs.

    Args:
        workspace_config: The configuration of the workspace.

    Returns:
        List of user IDs that should have access to the workspace.
        For private workspaces, returns specific user IDs.
        For public workspaces, returns ['*'] to indicate all users.
    """
    if workspace_config.get('private', False):
        user_ids = []
        workspace_user_name_or_ids = workspace_config.get('allowed_users', [])
        all_users = global_user_state.get_all_users()
        all_user_ids = {user.id for user in all_users}
        all_user_map = collections.defaultdict(list)
        for user in all_users:
            all_user_map[user.name].append(user.id)

        # Resolve user names to IDs
        for user_name_or_id in workspace_user_name_or_ids:
            if user_name_or_id in all_user_ids:
                user_ids.append(user_name_or_id)
            elif user_name_or_id in all_user_map:
                if len(all_user_map[user_name_or_id]) > 1:
                    user_ids_str = ', '.join(all_user_map[user_name_or_id])
                    raise ValueError(
                        f'User {user_name_or_id!r} has multiple IDs: '
                        f'{user_ids_str}. Please specify the user '
                        f'ID instead.')
                user_ids.append(all_user_map[user_name_or_id][0])
            else:
                logger.warning(
                    f'User {user_name_or_id!r} not found in all users')
                continue
        return user_ids
    else:
        # Public workspace - return '*' to indicate all users should have access
        return ['*']



"""Utilities for streaming logs from response."""

import asyncio
import collections
import pathlib
from typing import AsyncGenerator, Deque, List, Optional

import aiofiles
import fastapi

from sky import sky_logging
from sky.server.requests import requests as requests_lib
from sky.utils import message_utils
from sky.utils import rich_utils

logger = sky_logging.init_logger(__name__)

# When streaming log lines, buffer the lines in memory and flush them in chunks
# to improve log tailing throughput. Buffer size is the max size bytes of each
# chunk and the timeout threshold for flushing the buffer to ensure
# responsiveness.
_BUFFER_SIZE = 8 * 1024  # 8KB
_BUFFER_TIMEOUT = 0.02  # 20ms
_HEARTBEAT_INTERVAL = 30


async def _yield_log_file_with_payloads_skipped(
        log_file) -> AsyncGenerator[str, None]:
    async for line in log_file:
        if not line:
            return
        is_payload, line_str = message_utils.decode_payload(
            line.decode('utf-8'), raise_for_mismatch=False)
        if is_payload:
            continue

        yield line_str


async def log_streamer(request_id: Optional[str],
                       log_path: pathlib.Path,
                       plain_logs: bool = False,
                       tail: Optional[int] = None,
                       follow: bool = True) -> AsyncGenerator[str, None]:
    """Streams the logs of a request.

    Args:
        request_id: The request ID to check whether the log tailing process
            should be stopped.
        log_path: The path to the log file.
        plain_logs: Whether to show plain logs.
        tail: The number of lines to tail. If None, tail the whole file.
        follow: Whether to follow the log file.
    """

    if request_id is not None:
        status_msg = rich_utils.EncodedStatusMessage(
            f'[dim]Checking request: {request_id}[/dim]')
        request_task = requests_lib.get_request(request_id)

        if request_task is None:
            raise fastapi.HTTPException(
                status_code=404, detail=f'Request {request_id} not found')
        request_id = request_task.request_id

        # Do not show the waiting spinner if the request is a fast, non-blocking
        # request.
        show_request_waiting_spinner = (not plain_logs and
                                        request_task.schedule_type
                                        == requests_lib.ScheduleType.LONG)

        if show_request_waiting_spinner:
            yield status_msg.init()
            yield status_msg.start()
        last_waiting_msg = ''
        waiting_msg = (f'Waiting for {request_task.name!r} request to be '
                       f'scheduled: {request_id}')
        while request_task.status < requests_lib.RequestStatus.RUNNING:
            if request_task.status_msg is not None:
                waiting_msg = request_task.status_msg
            if show_request_waiting_spinner:
                yield status_msg.update(f'[dim]{waiting_msg}[/dim]')
            elif plain_logs and waiting_msg != last_waiting_msg:
                # Only log when waiting message changes.
                last_waiting_msg = waiting_msg
                # Use smaller padding (1024 bytes) to force browser rendering
                yield f'{waiting_msg}' + ' ' * 4096 + '\n'
            # Sleep shortly to avoid storming the DB and CPU and allow other
            # coroutines to run. This busy waiting loop is performance critical
            # for short-running requests, so we do not want to yield too long.
            await asyncio.sleep(0.1)
            request_task = requests_lib.get_request(request_id)
            if not follow:
                break
        if show_request_waiting_spinner:
            yield status_msg.stop()

    async with aiofiles.open(log_path, 'rb') as f:
        async for chunk in _tail_log_file(f, request_id, plain_logs, tail,
                                          follow):
            yield chunk


async def _tail_log_file(f: aiofiles.threadpool.binary.AsyncBufferedReader,
                         request_id: Optional[str] = None,
                         plain_logs: bool = False,
                         tail: Optional[int] = None,
                         follow: bool = True) -> AsyncGenerator[str, None]:
    """Tail the opened log file, buffer the lines and flush in chunks."""

    if tail is not None:
        # Find last n lines of the log file. Do not read the whole file into
        # memory.
        # TODO(zhwu): this will include the control lines for rich status,
        # which may not lead to exact tail lines when showing on the client
        # side.
        lines: Deque[str] = collections.deque(maxlen=tail)
        async for line_str in _yield_log_file_with_payloads_skipped(f):
            lines.append(line_str)
        for line_str in lines:
            yield line_str

    last_heartbeat_time = asyncio.get_event_loop().time()

    # Buffer the lines in memory and flush them in chunks to improve log
    # tailing throughput.
    buffer: List[str] = []
    buffer_bytes = 0
    last_flush_time = asyncio.get_event_loop().time()

    async def flush_buffer() -> AsyncGenerator[str, None]:
        nonlocal buffer, buffer_bytes, last_flush_time
        if buffer:
            yield ''.join(buffer)
            buffer.clear()
            buffer_bytes = 0
            last_flush_time = asyncio.get_event_loop().time()

    while True:
        # Sleep 0 to yield control to allow other coroutines to run,
        # while keeps the loop tight to make log stream responsive.
        await asyncio.sleep(0)
        current_time = asyncio.get_event_loop().time()
        # Flush the buffer when it is not empty and the buffer is full or the
        # flush timeout is reached.
        if buffer and (buffer_bytes >= _BUFFER_SIZE or
                       (current_time - last_flush_time) >= _BUFFER_TIMEOUT):
            async for chunk in flush_buffer():
                yield chunk

        line: Optional[bytes] = await f.readline()
        if not line:
            if request_id is not None:
                request_task = requests_lib.get_request(request_id)
                if request_task.status > requests_lib.RequestStatus.RUNNING:
                    if (request_task.status ==
                            requests_lib.RequestStatus.CANCELLED):
                        if request_task.should_retry:
                            buffer.append(
                                message_utils.encode_payload(
                                    rich_utils.Control.RETRY.encode('')))
                        else:
                            buffer.append(
                                f'{request_task.name!r} request {request_id}'
                                ' cancelled\n')
                    break
            if not follow:
                break

            if current_time - last_heartbeat_time >= _HEARTBEAT_INTERVAL:
                # Currently just used to keep the connection busy, refer to
                # https://github.com/skypilot-org/skypilot/issues/5750 for
                # more details.
                buffer.append(
                    message_utils.encode_payload(
                        rich_utils.Control.HEARTBEAT.encode('')))
                last_heartbeat_time = current_time

            # Sleep shortly to avoid storming the DB and CPU, this has
            # little impact on the responsivness here since we are waiting
            # for a new line to come in.
            await asyncio.sleep(0.1)
            continue

        # Refresh the heartbeat time, this is a trivial optimization for
        # performance but it helps avoid unnecessary heartbeat strings
        # being printed when the client runs in an old version.
        last_heartbeat_time = asyncio.get_event_loop().time()
        line_str = line.decode('utf-8')
        if plain_logs:
            is_payload, line_str = message_utils.decode_payload(
                line_str, raise_for_mismatch=False)
            # TODO(aylei): implement heartbeat mechanism for plain logs,
            # sending invisible characters might be okay.
            if is_payload:
                continue
        buffer.append(line_str)
        buffer_bytes += len(line_str.encode('utf-8'))

    # Flush remaining lines in the buffer.
    async for chunk in flush_buffer():
        yield chunk


def stream_response(
    request_id: str, logs_path: pathlib.Path,
    background_tasks: fastapi.BackgroundTasks
) -> fastapi.responses.StreamingResponse:

    async def on_disconnect():
        logger.info(f'User terminated the connection for request '
                    f'{request_id}')
        requests_lib.kill_requests([request_id])

    # The background task will be run after returning a response.
    # https://fastapi.tiangolo.com/tutorial/background-tasks/
    background_tasks.add_task(on_disconnect)

    return fastapi.responses.StreamingResponse(
        log_streamer(request_id, logs_path),
        media_type='text/plain',
        headers={
            'Cache-Control': 'no-cache, no-transform',
            'X-Accel-Buffering': 'no',
            'Transfer-Encoding': 'chunked'
        })



"""Constants for the database schemas."""

import contextlib
import logging
import os
import pathlib

from alembic import command as alembic_command
from alembic.config import Config
from alembic.runtime import migration
import filelock
import sqlalchemy

from sky import sky_logging
from sky.skylet import constants

logger = sky_logging.init_logger(__name__)

DB_INIT_LOCK_TIMEOUT_SECONDS = 10

GLOBAL_USER_STATE_DB_NAME = 'state_db'
GLOBAL_USER_STATE_VERSION = '006'
GLOBAL_USER_STATE_LOCK_PATH = '~/.sky/locks/.state_db.lock'

SPOT_JOBS_DB_NAME = 'spot_jobs_db'
SPOT_JOBS_VERSION = '003'
SPOT_JOBS_LOCK_PATH = '~/.sky/locks/.spot_jobs_db.lock'

SERVE_DB_NAME = 'serve_db'
SERVE_VERSION = '001'
SERVE_LOCK_PATH = '~/.sky/locks/.serve_db.lock'


def get_engine(db_name: str):
    conn_string = None
    if os.environ.get(constants.ENV_VAR_IS_SKYPILOT_SERVER) is not None:
        conn_string = os.environ.get(constants.ENV_VAR_DB_CONNECTION_URI)
    if conn_string:
        engine = sqlalchemy.create_engine(conn_string,
                                          poolclass=sqlalchemy.NullPool)
    else:
        db_path = os.path.expanduser(f'~/.sky/{db_name}.db')
        pathlib.Path(db_path).parents[0].mkdir(parents=True, exist_ok=True)
        engine = sqlalchemy.create_engine('sqlite:///' + db_path)
    return engine


@contextlib.contextmanager
def db_lock(db_name: str):
    lock_path = os.path.expanduser(f'~/.sky/locks/.{db_name}.lock')
    try:
        with filelock.FileLock(lock_path, timeout=DB_INIT_LOCK_TIMEOUT_SECONDS):
            yield
    except filelock.Timeout as e:
        raise RuntimeError(f'Failed to initialize database due to a timeout '
                           f'when trying to acquire the lock at '
                           f'{lock_path}. '
                           'Please try again or manually remove the lock '
                           f'file if you believe it is stale.') from e


def get_alembic_config(engine: sqlalchemy.engine.Engine, section: str):
    """Get Alembic configuration for the given section"""
    # From sky/utils/db/migration_utils.py -> sky/setup_files/alembic.ini
    alembic_ini_path = os.path.join(
        os.path.dirname(os.path.dirname(os.path.dirname(__file__))),
        'setup_files', 'alembic.ini')
    alembic_cfg = Config(alembic_ini_path, ini_section=section)

    # Override the database URL to match SkyPilot's current connection
    # Use render_as_string to get the full URL with password
    url = engine.url.render_as_string(hide_password=False)
    # Replace % with %% to escape the % character in the URL
    # set_section_option uses variable interpolation, which treats % as a
    # special character.
    # any '%' symbol not used for interpolation needs to be escaped.
    url = url.replace('%', '%%')
    alembic_cfg.set_section_option(section, 'sqlalchemy.url', url)

    return alembic_cfg


def needs_upgrade(engine: sqlalchemy.engine.Engine, section: str,
                  target_revision: str):
    """Check if the database needs to be upgraded.

    Args:
        engine: SQLAlchemy engine for the database
        section: Alembic section to upgrade (e.g., 'state_db' or 'spot_jobs_db')
        target_revision: Target revision to upgrade to (e.g., '001')
    """
    current_rev = None

    # get alembic config for the given section
    alembic_config = get_alembic_config(engine, section)
    version_table = alembic_config.get_section_option(
        alembic_config.config_ini_section, 'version_table', 'alembic_version')

    with engine.connect() as connection:
        context = migration.MigrationContext.configure(
            connection, opts={'version_table': version_table})
        current_rev = context.get_current_revision()

    if current_rev is None:
        return True

    # Compare revisions - assuming they are numeric strings like '001', '002'
    current_rev_num = int(current_rev)
    target_rev_num = int(target_revision)

    return current_rev_num < target_rev_num


def safe_alembic_upgrade(engine: sqlalchemy.engine.Engine, section: str,
                         target_revision: str):
    """Upgrade the database if needed. Uses a file lock to ensure
    that only one process tries to upgrade the database at a time.

    Args:
        engine: SQLAlchemy engine for the database
        section: Alembic section to upgrade (e.g., 'state_db' or 'spot_jobs_db')
        target_revision: Target revision to upgrade to (e.g., '001')
    """
    # set alembic logger to warning level
    alembic_logger = logging.getLogger('alembic')
    alembic_logger.setLevel(logging.WARNING)

    alembic_config = get_alembic_config(engine, section)

    # only acquire lock if db needs upgrade
    if needs_upgrade(engine, section, target_revision):
        with db_lock(section):
            # check again if db needs upgrade in case another
            # process upgraded it while we were waiting for the lock
            if needs_upgrade(engine, section, target_revision):
                alembic_command.upgrade(alembic_config, target_revision)



"""Scheduler for managed jobs.

Once managed jobs are submitted via submit_job, the scheduler is responsible for
the business logic of deciding when they are allowed to start, and choosing the
right one to start. The scheduler will also schedule jobs that are already live
but waiting to launch a new task or recover.

The scheduler is not its own process - instead, maybe_schedule_next_jobs() can
be called from any code running on the managed jobs controller instance to
trigger scheduling of new jobs if possible. This function should be called
immediately after any state change that could result in jobs newly being able to
be scheduled. If the job is running in a pool, the scheduler will only schedule
jobs for the same pool, because the resources limitations are per-pool (see the
following section for more details).

The scheduling logic limits #running jobs according to three limits:
1. The number of jobs that can be launching (that is, STARTING or RECOVERING) at
   once, based on the number of CPUs. This the most compute-intensive part of
   the job lifecycle, which is why we have an additional limit.
   See sky/utils/controller_utils.py::_get_launch_parallelism.
2. The number of jobs that can be running at any given time, based on the amount
   of memory. Since the job controller is doing very little once a job starts
   (just checking its status periodically), the most significant resource it
   consumes is memory.
   See sky/utils/controller_utils.py::_get_job_parallelism.
3. The number of jobs that can be running in a pool at any given time, based on
   the number of ready workers in the pool. (See _can_start_new_job.)

The state of the scheduler is entirely determined by the schedule_state column
of all the jobs in the job_info table. This column should only be modified via
the functions defined in this file. We will always hold the lock while modifying
this state. See state.ManagedJobScheduleState.

Nomenclature:
- job: same as managed job (may include multiple tasks)
- launch/launching: launching a cluster (sky.launch) as part of a job
- start/run: create the job controller process for a job
- schedule: transition a job to the LAUNCHING state, whether a new job or a job
  that is already alive
- alive: a job controller exists (includes multiple schedule_states: ALIVE,
  ALIVE_WAITING, LAUNCHING)
"""

from argparse import ArgumentParser
import contextlib
import os
import sys
import time
from typing import Optional

import filelock

from sky import exceptions
from sky import sky_logging
from sky.jobs import constants as managed_job_constants
from sky.jobs import state
from sky.serve import serve_utils
from sky.skylet import constants
from sky.utils import common_utils
from sky.utils import controller_utils
from sky.utils import subprocess_utils

logger = sky_logging.init_logger('sky.jobs.controller')

_ALIVE_JOB_LAUNCH_WAIT_INTERVAL = 0.5


def _start_controller(job_id: int, dag_yaml_path: str, env_file_path: str,
                      pool: Optional[str]) -> None:
    activate_python_env_cmd = (f'{constants.ACTIVATE_SKY_REMOTE_PYTHON_ENV};')
    source_environment_cmd = (f'source {env_file_path};'
                              if env_file_path else '')
    maybe_pool_arg = (f'--pool {pool}' if pool is not None else '')
    run_controller_cmd = (
        f'{sys.executable} -u -m sky.jobs.controller '
        f'{dag_yaml_path} --job-id {job_id} {maybe_pool_arg};')

    # If the command line here is changed, please also update
    # utils._controller_process_alive. The substring `--job-id X`
    # should be in the command.
    run_cmd = (f'{activate_python_env_cmd}'
               f'{source_environment_cmd}'
               f'{run_controller_cmd}')

    logs_dir = os.path.expanduser(
        managed_job_constants.JOBS_CONTROLLER_LOGS_DIR)
    os.makedirs(logs_dir, exist_ok=True)
    log_path = os.path.join(logs_dir, f'{job_id}.log')

    pid = subprocess_utils.launch_new_process_tree(run_cmd, log_output=log_path)
    state.set_job_controller_pid(job_id, pid)

    logger.debug(f'Job {job_id} started with pid {pid}')


def maybe_schedule_next_jobs() -> None:
    """Determine if any managed jobs can be scheduled, and if so, schedule them.

    Here, "schedule" means to select job that is waiting, and allow it to
    proceed. It does NOT mean to submit a job to the scheduler.

    For newly submitted jobs, scheduling means updating the state of the jobs,
    and starting the job controller process. For jobs that are already alive but
    are waiting to launch a new task or recover, just update the state of the
    job to indicate that the launch can proceed.

    This function transitions jobs into LAUNCHING on a best-effort basis. That
    is, if we can start any jobs, we will, but if not, we will exit (almost)
    immediately. It's expected that if some WAITING or ALIVE_WAITING jobs cannot
    be started now (either because the lock is held, or because there are not
    enough resources), another call to this function will be made whenever that
    situation is resolved. (If the lock is held, the lock holder should start
    the jobs. If there aren't enough resources, the next controller to exit and
    free up resources should start the jobs.)

    If this function obtains the lock, it will launch as many jobs as possible
    before releasing the lock. This is what allows other calls to exit
    immediately if the lock is held, while ensuring that all jobs are started as
    soon as possible.

    This uses subprocess_utils.launch_new_process_tree() to start the controller
    processes, which should be safe to call from pretty much any code running on
    the jobs controller instance. New job controller processes will be detached
    from the current process and there will not be a parent/child relationship.
    See launch_new_process_tree for more.

    After adding the pool support, this function will be called in a per-pool
    basis. We employ resources limitation for each pool given the number of
    ready workers in the pool. Each pool will have its own scheduler queue,
    indicating by the argument `pool`. Finished job in pool 1 will only trigger
    another jobs in pool 1, but the job in pool 2 will still be waiting. When
    the `pool` argument is None, it schedules a job regardless of the pool.
    """
    try:
        # We must use a global lock rather than a per-job lock to ensure correct
        # parallelism control. If we cannot obtain the lock, exit immediately.
        # The current lock holder is expected to launch any jobs it can before
        # releasing the lock.
        with filelock.FileLock(controller_utils.get_resources_lock_path(),
                               blocking=False):
            while True:
                maybe_next_job = state.get_waiting_job()
                if maybe_next_job is None:
                    # Nothing left to start, break from scheduling loop
                    break
                actual_pool = maybe_next_job['pool']

                current_state = maybe_next_job['schedule_state']

                assert current_state in (
                    state.ManagedJobScheduleState.ALIVE_WAITING,
                    state.ManagedJobScheduleState.WAITING), maybe_next_job

                # Note: we expect to get ALIVE_WAITING jobs before WAITING jobs,
                # since they will have been submitted and therefore started
                # first. The requirements to launch in an alive job are more
                # lenient, so there is no way that we wouldn't be able to launch
                # an ALIVE_WAITING job, but we would be able to launch a WAITING
                # job.
                if current_state == state.ManagedJobScheduleState.ALIVE_WAITING:
                    if not controller_utils.can_provision():
                        # Can't schedule anything, break from scheduling loop.
                        break
                elif current_state == state.ManagedJobScheduleState.WAITING:
                    if not _can_start_new_job(actual_pool):
                        # Can't schedule anything, break from scheduling loop.
                        break

                logger.debug(f'Scheduling job {maybe_next_job["job_id"]}')
                state.scheduler_set_launching(maybe_next_job['job_id'],
                                              current_state)

                if current_state == state.ManagedJobScheduleState.WAITING:
                    # The job controller has not been started yet. We must start
                    # it.

                    job_id = maybe_next_job['job_id']
                    dag_yaml_path = maybe_next_job['dag_yaml_path']
                    env_file_path = maybe_next_job['env_file_path']

                    _start_controller(job_id, dag_yaml_path, env_file_path,
                                      actual_pool)

    except filelock.Timeout:
        # If we can't get the lock, just exit. The process holding the lock
        # should launch any pending jobs.
        pass


def submit_job(job_id: int, dag_yaml_path: str, original_user_yaml_path: str,
               env_file_path: str, priority: int, pool: Optional[str]) -> None:
    """Submit an existing job to the scheduler.

    This should be called after a job is created in the `spot` table as
    PENDING. It will tell the scheduler to try and start the job controller, if
    there are resources available. It may block to acquire the lock, so it
    should not be on the critical path for `sky jobs launch -d`.

    The user hash should be set (e.g. via SKYPILOT_USER_ID) before calling this.
    """
    with filelock.FileLock(controller_utils.get_resources_lock_path()):
        is_resume = state.scheduler_set_waiting(job_id, dag_yaml_path,
                                                original_user_yaml_path,
                                                env_file_path,
                                                common_utils.get_user_hash(),
                                                priority)
    if is_resume:
        _start_controller(job_id, dag_yaml_path, env_file_path, pool)
    else:
        maybe_schedule_next_jobs()


@contextlib.contextmanager
def scheduled_launch(job_id: int):
    """Launch as part of an ongoing job.

    A newly started job will already be LAUNCHING, and this will immediately
    enter the context.

    If a job is ongoing (ALIVE schedule_state), there are two scenarios where we
    may need to call sky.launch again during the course of a job controller:
    - for tasks after the first task
    - for recovery

    This function will mark the job as ALIVE_WAITING, which indicates to the
    scheduler that it wants to transition back to LAUNCHING. Then, it will wait
    until the scheduler transitions the job state, before entering the context.

    On exiting the context, the job will transition to ALIVE.

    This should only be used within the job controller for the given job_id. If
    multiple uses of this context are nested, behavior is undefined. Don't do
    that.
    """
    pool = state.get_pool_from_job_id(job_id)
    # For pool, since there is no execution.launch, we don't need to have all
    # the ALIVE_WAITING state. The state transition will be
    # WAITING -> ALIVE -> DONE without any intermediate transitions.
    if pool is not None:
        yield
        return

    # If we're already in LAUNCHING schedule_state, we don't need to wait.
    # This may be the case for the first launch of a job.
    if (state.get_job_schedule_state(job_id) !=
            state.ManagedJobScheduleState.LAUNCHING):
        # Since we aren't LAUNCHING, we need to wait to be scheduled.
        _set_alive_waiting(job_id)

        while (state.get_job_schedule_state(job_id) !=
               state.ManagedJobScheduleState.LAUNCHING):
            time.sleep(_ALIVE_JOB_LAUNCH_WAIT_INTERVAL)

    try:
        yield
    except exceptions.NoClusterLaunchedError:
        # NoClusterLaunchedError is indicates that the job is in retry backoff.
        # We should transition to ALIVE_BACKOFF instead of ALIVE.
        with filelock.FileLock(controller_utils.get_resources_lock_path()):
            state.scheduler_set_alive_backoff(job_id)
        raise
    else:
        with filelock.FileLock(controller_utils.get_resources_lock_path()):
            state.scheduler_set_alive(job_id)
    finally:
        maybe_schedule_next_jobs()


def job_done(job_id: int, idempotent: bool = False) -> None:
    """Transition a job to DONE.

    If idempotent is True, this will not raise an error if the job is already
    DONE.

    The job could be in any terminal ManagedJobStatus. However, once DONE, it
    should never transition back to another state.
    """
    if idempotent and (state.get_job_schedule_state(job_id)
                       == state.ManagedJobScheduleState.DONE):
        return

    with filelock.FileLock(controller_utils.get_resources_lock_path()):
        state.scheduler_set_done(job_id, idempotent)
    maybe_schedule_next_jobs()


def _set_alive_waiting(job_id: int) -> None:
    """Should use wait_until_launch_okay() to transition to this state."""
    with filelock.FileLock(controller_utils.get_resources_lock_path()):
        state.scheduler_set_alive_waiting(job_id)
    maybe_schedule_next_jobs()


def _can_start_new_job(pool: Optional[str]) -> bool:
    # Check basic resource limits
    # Pool jobs don't need to provision resources, so we skip the check.
    if not ((controller_utils.can_provision() or pool is not None) and
            controller_utils.can_start_new_process()):
        return False

    # Check if there are available workers in the pool
    if pool is not None:
        alive_jobs_in_pool = state.get_num_alive_jobs(pool)
        if alive_jobs_in_pool >= len(serve_utils.get_ready_replicas(pool)):
            logger.debug(f'No READY workers available in pool {pool}')
            return False

    return True


if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument('dag_yaml',
                        type=str,
                        help='The path to the user job yaml file.')
    parser.add_argument('--user-yaml-path',
                        type=str,
                        help='The path to the original user job yaml file.')
    parser.add_argument('--job-id',
                        required=True,
                        type=int,
                        help='Job id for the controller job.')
    parser.add_argument('--env-file',
                        type=str,
                        help='The path to the controller env file.')
    parser.add_argument('--pool',
                        type=str,
                        required=False,
                        default=None,
                        help='The pool to use for the controller job.')
    parser.add_argument(
        '--priority',
        type=int,
        default=constants.DEFAULT_PRIORITY,
        help=
        f'Job priority ({constants.MIN_PRIORITY} to {constants.MAX_PRIORITY}).'
        f' Default: {constants.DEFAULT_PRIORITY}.')
    args = parser.parse_args()
    submit_job(args.job_id, args.dag_yaml, args.user_yaml_path, args.env_file,
               args.priority, args.pool)



"""Alembic environment configuration for state database migrations."""
from logging.config import fileConfig

from alembic import context
from sqlalchemy import engine_from_config
from sqlalchemy import pool

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# NOTE: We intentionally disable Alembic's logging configuration to prevent
# it from overriding SkyPilot's logging setup. Alembic's fileConfig() call
# globally reconfigures Python's logging system, which can suppress SkyPilot's
# output messages that tests expect to see.
#
# Original code (now disabled):
if config.config_file_name is not None:
    fileConfig(config.config_file_name, disable_existing_loggers=False)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
target_metadata = None

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option('sqlalchemy.url')
    version_table = config.get_section_option(config.config_ini_section,
                                              'version_table',
                                              'alembic_version')
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={'paramstyle': 'named'},
        version_table=version_table,
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    connectable = engine_from_config(
        config.get_section(config.config_ini_section, {}),
        prefix='sqlalchemy.',
        poolclass=pool.NullPool,
    )
    version_table = config.get_section_option(config.config_ini_section,
                                              'version_table',
                                              'alembic_version')
    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            version_table=version_table,
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



"""Volume types and access modes."""
from typing import Any, Dict, Optional

from sky.utils import common_utils
from sky.utils import infra_utils
from sky.utils import resources_utils
from sky.utils import schemas


class Volume:
    """Volume specification."""

    def __init__(
            self,
            name: Optional[str] = None,
            type: Optional[str] = None,  # pylint: disable=redefined-builtin
            infra: Optional[str] = None,
            size: Optional[str] = None,
            resource_name: Optional[str] = None,
            config: Optional[Dict[str, Any]] = None):
        """Initialize a Volume instance.

        Args:
            name: Volume name
            type: Volume type (e.g., 'k8s-pvc')
            infra: Infrastructure specification
            size: Volume size
            config: Additional configuration
        """
        self.name = name
        self.type = type
        self.infra = infra
        self.size = size
        self.resource_name = resource_name
        self.config = config or {}

        self.cloud: Optional[str] = None
        self.region: Optional[str] = None
        self.zone: Optional[str] = None

    @classmethod
    def from_dict(cls, config_dict: Dict[str, Any]) -> 'Volume':
        """Create a Volume instance from a dictionary."""
        return cls(name=config_dict.get('name'),
                   type=config_dict.get('type'),
                   infra=config_dict.get('infra'),
                   size=config_dict.get('size'),
                   resource_name=config_dict.get('resource_name'),
                   config=config_dict.get('config', {}))

    def to_dict(self) -> Dict[str, Any]:
        """Convert the Volume to a dictionary."""
        return {
            'name': self.name,
            'type': self.type,
            'infra': self.infra,
            'size': self.size,
            'resource_name': self.resource_name,
            'config': self.config,
            'cloud': self.cloud,
            'region': self.region,
            'zone': self.zone,
        }

    def normalize_config(
            self,
            name: Optional[str] = None,
            infra: Optional[str] = None,
            type: Optional[str] = None,  # pylint: disable=redefined-builtin
            size: Optional[str] = None) -> None:
        """Override the volume config with CLI options,
           adjust and validate the config.

        Args:
            name: Volume name to override
            infra: Infrastructure to override
            type: Volume type to override
            size: Volume size to override
        """
        if name is not None:
            self.name = name
        if infra is not None:
            self.infra = infra
        if type is not None:
            self.type = type
        if size is not None:
            self.size = size

        # Validate schema
        common_utils.validate_schema(self.to_dict(),
                                     schemas.get_volume_schema(),
                                     'Invalid volumes config: ')

        # Adjust the volume config (e.g., parse size)
        self._adjust_config()

        # Validate the volume config
        self._validate_config()

        # Resolve the infrastructure options to cloud, region, zone
        infra_info = infra_utils.InfraInfo.from_str(self.infra)
        self.cloud = infra_info.cloud
        self.region = infra_info.region
        self.zone = infra_info.zone

    def _adjust_config(self) -> None:
        """Adjust the volume config (e.g., parse size)."""
        if self.size is None:
            return
        try:
            size = resources_utils.parse_memory_resource(self.size,
                                                         'size',
                                                         allow_rounding=True)
            if size == '0':
                raise ValueError('Size must be no less than 1Gi')
            self.size = size
        except ValueError as e:
            raise ValueError(f'Invalid size {self.size}: {e}') from e

    def _validate_config(self) -> None:
        """Validate the volume config."""
        if not self.resource_name and not self.size:
            raise ValueError('Size is required for new volumes. '
                             'Please specify the size in the YAML file or '
                             'use the --size flag.')



"""add workspace column to cluster_history table

Revision ID: 002
Revises: 001
Create Date: 2025-08-06

"""
# pylint: disable=invalid-name
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa

from sky.utils.db import db_utils

# revision identifiers, used by Alembic.
revision: str = '002'
down_revision: Union[str, Sequence[str], None] = '001'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade schema."""
    with op.get_context().autocommit_block():
        db_utils.add_column_to_table_alembic('cluster_history',
                                             'workspace',
                                             sa.Text(),
                                             server_default=None)
    pass


def downgrade() -> None:
    """Downgrade schema."""
    pass



from unittest import mock

from click import testing as cli_testing
import requests

from sky import clouds
from sky import server
from sky.client.cli import command
from sky.utils import status_lib

CLOUDS_TO_TEST = [
    'aws', 'gcp', 'ibm', 'azure', 'lambda', 'scp', 'oci', 'vsphere', 'nebius'
]


def mock_server_api_version(monkeypatch, version):
    original_request = requests.request

    def mock_request(method, url, *args, **kwargs):
        if '/api/health' in url and method == 'GET':
            mock_response = mock.MagicMock()
            mock_response.status_code = 200
            mock_response.json.return_value = {
                'api_version': version,
                'version': '1.0.0',
                'commit': '1234567890'
            }
            return mock_response
        return original_request(method, url, *args, **kwargs)

    monkeypatch.setattr(requests, 'request', mock_request)


# Global API server mocking to prevent real network calls
def mock_api_server_calls(monkeypatch):
    """Mock all API server related calls to prevent real network requests."""
    # Mock API server status checks
    mock_api_info = mock.MagicMock()
    mock_api_info.status = 'healthy'
    mock_api_info.api_version = 1
    mock_api_info.version = '1.0.0'
    mock_api_info.commit = 'test_commit'
    mock_api_info.user = None
    mock_api_info.basic_auth_enabled = False

    monkeypatch.setattr('sky.server.common.get_api_server_status',
                        lambda *args, **kwargs: mock_api_info)

    # Mock other server functions that might trigger real calls
    monkeypatch.setattr('sky.server.common.check_server_healthy',
                        lambda *args, **kwargs: ('healthy', mock_api_info))
    monkeypatch.setattr('sky.server.common.get_server_url',
                        lambda *args, **kwargs: 'http://localhost:46580')
    monkeypatch.setattr('sky.server.common.is_api_server_local', lambda: True)

    # Mock make_authenticated_request to prevent real HTTP calls
    def mock_make_authenticated_request(method, path, *args, **kwargs):
        mock_response = mock.MagicMock()
        mock_response.status_code = 200
        mock_response.headers = {
            'X-Skypilot-Request-ID': 'test_request_id_12345'
        }
        mock_response.history = []

        # Mock different endpoints
        if '/api/health' in path:
            mock_response.json.return_value = {
                'status': 'healthy',
                'api_version': 1,
                'version': '1.0.0',
                'version_on_disk': '1.0.0',
                'commit': 'test_commit',
                'user': None,
                'basic_auth_enabled': False
            }
        else:
            # For other endpoints, return a simple success response
            mock_response.json.return_value = {'status': 'success'}
            # Set the text attribute for request ID parsing
            mock_response.text = 'test_request_id_12345'

        return mock_response

    monkeypatch.setattr('sky.server.common.make_authenticated_request',
                        mock_make_authenticated_request)

    # Mock high-level SDK functions to prevent API calls entirely
    monkeypatch.setattr('sky.client.sdk.check',
                        lambda *args, **kwargs: 'test_request_id')

    # Create mock accelerator items that have the _asdict method like dataclasses
    class MockAcceleratorItem:

        def __init__(self,
                     accelerator_name='V100',
                     accelerator_count=1,
                     cloud='AWS',
                     instance_type='p3.2xlarge',
                     device_memory=16.0,
                     cpu_count=8,
                     memory=61.0,
                     price=1.0,
                     spot_price=0.5,
                     region='us-east-1'):
            self.accelerator_name = accelerator_name
            self.accelerator_count = accelerator_count
            self.cloud = cloud
            self.instance_type = instance_type
            self.device_memory = device_memory
            self.cpu_count = cpu_count
            self.memory = memory
            self.price = price
            self.spot_price = spot_price
            self.region = region

        def _asdict(self):
            return {
                'accelerator_name': self.accelerator_name,
                'accelerator_count': self.accelerator_count,
                'cloud': self.cloud,
                'instance_type': self.instance_type,
                'device_memory': self.device_memory,
                'cpu_count': self.cpu_count,
                'memory': self.memory,
                'price': self.price,
                'spot_price': self.spot_price,
                'region': self.region
            }

    # Mock list_accelerators to return a request ID (for specific accelerator queries)
    monkeypatch.setattr('sky.client.sdk.list_accelerators',
                        lambda *args, **kwargs: 'accelerators_request_id')

    # Mock list_accelerator_counts to return a request ID (for --all queries)
    monkeypatch.setattr('sky.client.sdk.list_accelerator_counts',
                        lambda *args, **kwargs: 'accelerator_counts_request_id')

    # Mock stream_and_get to return appropriate data based on request ID
    def mock_stream_and_get(request_id):
        if 'accelerators_request_id' in str(request_id):
            # Return dictionary with GPU names as keys and lists of accelerator items as values
            return {
                'V100': [MockAcceleratorItem('V100', 1, 'AWS', 'p3.2xlarge')],
                'T4': [MockAcceleratorItem('T4', 1, 'GCP', 'n1-standard-4')]
            }
        elif 'accelerator_counts_request_id' in str(request_id):
            # Return dictionary with GPU names as keys and lists of quantities as values
            # This is used for --all flag (when accelerator_str is None)
            return {
                'V100': [1, 2, 4, 8],  # Available quantities
                'T4': [1, 2, 4],
                'A100': [1, 2, 4, 8],
                'H100': [1, 2, 4],
            }
        elif 'status_request_id' in str(request_id):
            return [{
                'name': 'test-cluster',
                'handle': None,
                'credentials': {},
                'status': status_lib.ClusterStatus.UP,
                'workspace': 'default',
                'autostop': -1,  # -1 means disabled
                'to_down': False,
                'launched_at': 1640995200,  # Some timestamp
                'last_use': 'sky launch',
                'user_hash': 'test_user_hash',
                'user_name': 'test_user',
                'resources_str': '1x t2.micro',
                'resources_str_full': '1x t2.micro',
                'infra': 'AWS/us-east-1'
            }]
        elif 'jobs_queue_request_id' in str(request_id):
            return [{'job_id': 1, 'status': 'RUNNING'}]
        else:
            return []

    monkeypatch.setattr('sky.client.sdk.stream_and_get', mock_stream_and_get)

    # Mock sdk.get to return appropriate data based on request ID
    def mock_get(request_id):
        if 'enabled_clouds_request_id' in str(request_id):
            return ['aws', 'gcp']  # Return list of cloud names
        elif 'workspaces_request_id' in str(request_id):
            return ['default', 'workspace1']
        elif 'serve_status_request_id' in str(request_id):
            return [{
                'name': 'service1',
                'status': 'RUNNING',
                'endpoint': 'http://localhost:8080'
            }]
        else:
            return []

    monkeypatch.setattr('sky.client.sdk.get', mock_get)
    monkeypatch.setattr('sky.client.sdk.enabled_clouds',
                        lambda *args, **kwargs: 'enabled_clouds_request_id')
    monkeypatch.setattr('sky.client.sdk.realtime_kubernetes_gpu_availability',
                        lambda *args, **kwargs: 'k8s_gpu_request_id')
    monkeypatch.setattr('sky.client.sdk.kubernetes_node_info',
                        lambda *args, **kwargs: 'k8s_node_request_id')

    # Mock requests.request as well for any direct usage
    def mock_requests_request(method, url, *args, **kwargs):
        mock_response = mock.MagicMock()
        mock_response.status_code = 200
        mock_response.headers = {
            'X-Skypilot-Request-ID': 'test_request_id_12345'
        }
        mock_response.text = 'test_request_id_12345'
        mock_response.json.return_value = {'status': 'success'}
        return mock_response

    monkeypatch.setattr('requests.request', mock_requests_request)
    monkeypatch.setattr('requests.get', mock_requests_request)
    monkeypatch.setattr('requests.post', mock_requests_request)

    # Mock usage tracking
    monkeypatch.setattr('sky.usage.usage_lib.messages.usage.set_internal',
                        lambda: None)

    # Mock the rest client to prevent real API calls
    def mock_rest_request(method, url, *args, **kwargs):
        mock_response = mock.MagicMock()
        mock_response.status_code = 200
        mock_response.headers = {
            'X-Skypilot-Request-ID': 'test_request_id_12345'
        }
        mock_response.text = 'test_request_id_12345'
        mock_response.json.return_value = {'status': 'success'}
        return mock_response

    monkeypatch.setattr('sky.server.rest.request', mock_rest_request)
    monkeypatch.setattr('sky.server.rest.request_without_retry',
                        mock_rest_request)


class TestWithNoCloudEnabled:

    def test_show_gpus(self, monkeypatch):
        """Tests `sky show-gpus` can be invoked (but not correctness).

        Tests below correspond to the following terminal commands, in order:

        -> sky show-gpus
        -> sky show-gpus --all
        -> sky show-gpus V100:4
        -> sky show-gpus :4
        -> sky show-gpus V100:0
        -> sky show-gpus V100:-2
        -> sky show-gpus --cloud aws --region us-west-1
        -> sky show-gpus --cloud lambda
        -> sky show-gpus --cloud lambda --all
        -> sky show-gpus V100:4 --cloud lambda
        -> sky show-gpus V100:4 --cloud lambda --all
        """
        mock_api_server_calls(monkeypatch)

        cli_runner = cli_testing.CliRunner()
        result = cli_runner.invoke(command.show_gpus, [])
        assert not result.exit_code

        result = cli_runner.invoke(command.show_gpus, ['--all'])
        assert not result.exit_code

        result = cli_runner.invoke(command.show_gpus, ['V100:4'])
        assert not result.exit_code

        result = cli_runner.invoke(command.show_gpus, [':4'])
        assert not result.exit_code

        result = cli_runner.invoke(command.show_gpus, ['V100:0'])
        assert isinstance(result.exception, SystemExit)

        result = cli_runner.invoke(command.show_gpus, ['V100:-2'])
        assert isinstance(result.exception, SystemExit)

        result = cli_runner.invoke(command.show_gpus,
                                   ['--cloud', 'aws', '--region', 'us-west-1'])
        assert not result.exit_code

        result = cli_runner.invoke(command.show_gpus,
                                   ['--infra', 'aws/us-west-1'])
        assert not result.exit_code

        for cloud in CLOUDS_TO_TEST:
            result = cli_runner.invoke(command.show_gpus, ['--infra', cloud])
            assert not result.exit_code

            result = cli_runner.invoke(command.show_gpus,
                                       ['--cloud', cloud, '--all'])
            assert not result.exit_code

            result = cli_runner.invoke(command.show_gpus,
                                       ['V100', '--cloud', cloud])
            assert not result.exit_code

            result = cli_runner.invoke(command.show_gpus,
                                       ['V100:4', '--cloud', cloud])
            assert not result.exit_code

            result = cli_runner.invoke(command.show_gpus,
                                       ['V100:4', '--cloud', cloud, '--all'])
            assert isinstance(result.exception, SystemExit)

    def test_k8s_alias_check(self, monkeypatch):
        mock_api_server_calls(monkeypatch)

        cli_runner = cli_testing.CliRunner()

        result = cli_runner.invoke(command.check, ['k8s'])
        assert not result.exit_code

        result = cli_runner.invoke(command.check, ['kubernetes'])
        assert not result.exit_code

        # With our mocking, invalid cloud names won't raise ValueError
        # since we're bypassing the real validation logic
        result = cli_runner.invoke(command.check, ['notarealcloud'])
        # In mocked environment, this should succeed rather than raise ValueError
        assert not result.exit_code


class TestHelperFunctions:

    def test_get_cluster_records_and_set_ssh_config(self, monkeypatch):
        """Tests _get_cluster_records_and_set_ssh_config with mocked components."""
        mock_api_server_calls(monkeypatch)

        # Mock cluster records that would be returned by stream_and_get
        mock_handle = mock.MagicMock()
        mock_handle.cluster_name = 'test-cluster'
        mock_handle.cached_external_ips = ['1.2.3.4']
        mock_handle.cached_external_ssh_ports = [22]
        mock_handle.docker_user = None
        mock_handle.ssh_user = 'ubuntu'
        mock_handle.launched_resources.cloud = mock.MagicMock()

        mock_records = [{
            'name': 'test-cluster',
            'handle': mock_handle,
            'credentials': {
                'ssh_user': 'ubuntu',
                'ssh_private_key': '/path/to/key.pem'
            }
        }]

        # Mock the SDK calls
        def mock_status(*args, **kwargs):
            return 'request-id'

        def mock_stream_and_get(*args, **kwargs):
            return mock_records

        monkeypatch.setattr('sky.client.sdk.status', mock_status)
        monkeypatch.setattr('sky.client.sdk.stream_and_get',
                            mock_stream_and_get)

        # Mock SSHConfigHelper methods
        mock_add_cluster = mock.MagicMock()
        mock_remove_cluster = mock.MagicMock()
        mock_list_cluster_names = mock.MagicMock(
            return_value=['test-cluster', 'old-cluster'])

        monkeypatch.setattr(
            'sky.utils.cluster_utils.SSHConfigHelper.add_cluster',
            mock_add_cluster)
        monkeypatch.setattr(
            'sky.utils.cluster_utils.SSHConfigHelper.remove_cluster',
            mock_remove_cluster)
        monkeypatch.setattr(
            'sky.utils.cluster_utils.SSHConfigHelper.list_cluster_names',
            mock_list_cluster_names)

        # Test case 1: Get records for specific clusters
        records = command._get_cluster_records_and_set_ssh_config(
            ['test-cluster'])
        assert records == mock_records
        mock_add_cluster.assert_called_once_with('test-cluster', ['1.2.3.4'], {
            'ssh_user': 'ubuntu',
            'ssh_private_key': '/path/to/key.pem'
        }, [22], None, 'ubuntu')
        # Shouldn't remove anything because all clusters provided are in the returned records
        mock_remove_cluster.assert_not_called()

        # Reset mocks for next test
        mock_add_cluster.reset_mock()
        mock_remove_cluster.reset_mock()

        # Test case 2: Get records for all users
        records = command._get_cluster_records_and_set_ssh_config(
            None, all_users=True)
        assert records == mock_records
        mock_add_cluster.assert_called_once()
        # Should remove old-cluster since it's not in the returned records
        mock_remove_cluster.assert_called_once_with('old-cluster')

        # Test case 3: Test with a cluster that has no handle/credentials
        mock_records_no_handle = [{'name': 'test-cluster', 'handle': None}]
        monkeypatch.setattr('sky.client.sdk.stream_and_get',
                            lambda *args, **kwargs: mock_records_no_handle)

        records = command._get_cluster_records_and_set_ssh_config(
            ['test-cluster'])
        assert records == mock_records_no_handle
        # Should remove the cluster from SSH config since it has no handle
        mock_remove_cluster.assert_called_with('test-cluster')

        # Reset mocks for next test
        mock_add_cluster.reset_mock()
        mock_remove_cluster.reset_mock()

        # Test case 4: Test with a cluster that is using kubernetes
        mock_k8s_handle = mock.MagicMock()
        mock_k8s_handle.cluster_name = 'test-cluster'
        mock_k8s_handle.cached_external_ips = ['1.2.3.4']
        mock_k8s_handle.cached_external_ssh_ports = [22]
        mock_k8s_handle.docker_user = None
        mock_k8s_handle.ssh_user = 'ubuntu'
        mock_k8s_handle.launched_resources.cloud = clouds.Kubernetes()
        mock_records_kubernetes = [{
            'name': 'test-cluster',
            'handle': mock_k8s_handle,
            'credentials': {
                'ssh_user': 'ubuntu',
                'ssh_private_key': '/path/to/key.pem'
            },
        }]
        server_url = 'http://localhost:8000'
        monkeypatch.setattr('sky.server.common.get_server_url',
                            lambda: server_url)
        monkeypatch.setattr('sky.client.sdk.stream_and_get',
                            lambda *args, **kwargs: mock_records_kubernetes)
        records = command._get_cluster_records_and_set_ssh_config(
            ['test-cluster'])
        assert records == mock_records_kubernetes
        mock_add_cluster.assert_called_once()
        added_cluster_args = mock_add_cluster.call_args
        assert added_cluster_args[0][0] == 'test-cluster'
        assert added_cluster_args[0][1] == ['1.2.3.4']
        # proxy command should be set, but is dependent on the server url, so we don't check the exact value
        assert added_cluster_args[0][2].get('ssh_proxy_command') is not None
        assert server_url in added_cluster_args[0][2].get('ssh_proxy_command')
        assert added_cluster_args[0][2].get(
            'ssh_private_key') == '/path/to/key.pem'
        assert added_cluster_args[0][2].get('ssh_user') == 'ubuntu'
        assert added_cluster_args[0][3] == [22]
        assert added_cluster_args[0][4] is None
        assert added_cluster_args[0][5] == 'ubuntu'
        mock_remove_cluster.assert_not_called()

    def test_list_to_str_float_formatting(self):
        """Test that _list_to_str formats whole number floats as integers.
        
        Regression test for GitHub issue #6484 where requestable quantities
        were shown as '1.0, 2.0, 4.0, 8.0' instead of '1, 2, 4, 8'.
        """

        # Define the function locally to test it (since it's nested in show_gpus)
        def _list_to_str(lst):

            def format_number(e):
                # If it's a float that's a whole number, display as int
                if isinstance(e, float) and e.is_integer():
                    return str(int(e))
                return str(e)

            return ', '.join([format_number(e) for e in lst])

        # Test case from GitHub issue #6484: whole number floats should display as integers
        assert _list_to_str([1.0, 2.0, 4.0, 8.0]) == '1, 2, 4, 8'

        # Test mixed floats and integers
        assert _list_to_str([1, 2.0, 4.0, 8]) == '1, 2, 4, 8'

        # Test fractional floats should remain as floats
        assert _list_to_str([1.5, 2.0, 4.0]) == '1.5, 2, 4'

        # Test edge cases
        assert _list_to_str([0.0, 1.0]) == '0, 1'
        assert _list_to_str([0.5, 1.0, 2.5]) == '0.5, 1, 2.5'

        # Test integers remain unchanged
        assert _list_to_str([1, 2, 4, 8]) == '1, 2, 4, 8'

        # Test empty list
        assert _list_to_str([]) == ''

        # Test single element
        assert _list_to_str([1.0]) == '1'
        assert _list_to_str([1.5]) == '1.5'

    def test_show_gpus_k8s_float_formatting(self, monkeypatch):
        """Integration test for sky show-gpus --infra k8s output formatting.
        
        Regression test for GitHub issue #6484 to ensure that requestable quantities
        are displayed as integers (1, 2, 4, 8) instead of floats (1.0, 2.0, 4.0, 8.0).
        """
        mock_api_server_calls(monkeypatch)

        # Import required modules
        from sky import models

        # Mock kubernetes GPU availability data with float counts (the bug scenario)
        mock_gpu_availability = [
            models.RealtimeGpuAvailability(
                gpu='H100_NVLINK_80GB',
                counts=[1.0, 2.0, 4.0,
                        8.0],  # These should be formatted as integers
                capacity=16,
                available=16)
        ]

        # Mock enabled clouds to include kubernetes
        def mock_get(request_id):
            if 'enabled_clouds_request_id' in str(request_id):
                return ['kubernetes']  # Enable kubernetes
            return []

        monkeypatch.setattr('sky.client.sdk.get', mock_get)
        monkeypatch.setattr('sky.client.sdk.enabled_clouds',
                            lambda *args, **kwargs: 'enabled_clouds_request_id')

        # Mock the kubernetes contexts and GPU availability
        monkeypatch.setattr('sky.clouds.Kubernetes.existing_allowed_contexts',
                            lambda: ['test-context'])
        monkeypatch.setattr(
            'sky.client.sdk.realtime_kubernetes_gpu_availability',
            lambda *args, **kwargs: 'k8s_gpu_request_id')
        monkeypatch.setattr('sky.client.sdk.kubernetes_node_info',
                            lambda *args, **kwargs: 'k8s_node_request_id')

        # Mock node info to avoid additional API calls
        mock_node_info = models.KubernetesNodesInfo(node_info_dict={},
                                                    hint=None)

        def mock_stream_and_get_conditional(request_id):
            if 'k8s_gpu' in request_id:
                return [('test-context', mock_gpu_availability)]
            elif 'k8s_node' in request_id:
                return mock_node_info
            return []

        monkeypatch.setattr('sky.client.sdk.stream_and_get',
                            mock_stream_and_get_conditional)

        # Run the command
        cli_runner = cli_testing.CliRunner()
        result = cli_runner.invoke(command.show_gpus, ['--infra', 'k8s'])

        # Check that command succeeded
        assert result.exit_code == 0, f"Command failed with output: {result.output}"

        # Check that the output contains properly formatted integers (not floats)
        output = result.output
        assert '1, 2, 4, 8' in output, f"Expected '1, 2, 4, 8' in output, got: {output}"
        # Ensure it doesn't contain the problematic float format
        assert '1.0, 2.0, 4.0, 8.0' not in output, f"Found float format in output: {output}"



"""Utilities for REST API."""
import asyncio
import contextlib
import dataclasses
import enum
import functools
import json
import os
import pathlib
import shutil
import signal
import sqlite3
import threading
import time
import traceback
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple

import colorama
import filelock

from sky import exceptions
from sky import global_user_state
from sky import sky_logging
from sky import skypilot_config
from sky.server import common as server_common
from sky.server import constants as server_constants
from sky.server import daemons
from sky.server.requests import payloads
from sky.server.requests.serializers import decoders
from sky.server.requests.serializers import encoders
from sky.utils import common_utils
from sky.utils import subprocess_utils
from sky.utils import ux_utils
from sky.utils.db import db_utils

logger = sky_logging.init_logger(__name__)

# Tables in task.db.
REQUEST_TABLE = 'requests'
COL_CLUSTER_NAME = 'cluster_name'
COL_USER_ID = 'user_id'
COL_STATUS_MSG = 'status_msg'
COL_SHOULD_RETRY = 'should_retry'
COL_FINISHED_AT = 'finished_at'
REQUEST_LOG_PATH_PREFIX = '~/sky_logs/api_server/requests'

DEFAULT_REQUESTS_RETENTION_HOURS = 24  # 1 day

# TODO(zhwu): For scalability, there are several TODOs:
# [x] Have a way to queue requests.
# [ ] Move logs to persistent place.
# [ ] Deploy API server in a autoscaling fashion.


class RequestStatus(enum.Enum):
    """The status of a request."""

    PENDING = 'PENDING'
    RUNNING = 'RUNNING'
    SUCCEEDED = 'SUCCEEDED'
    FAILED = 'FAILED'
    CANCELLED = 'CANCELLED'

    def __gt__(self, other):
        return (list(RequestStatus).index(self) >
                list(RequestStatus).index(other))

    def colored_str(self):
        color = _STATUS_TO_COLOR[self]
        return f'{color}{self.value}{colorama.Style.RESET_ALL}'

    @classmethod
    def finished_status(cls) -> List['RequestStatus']:
        return [cls.SUCCEEDED, cls.FAILED, cls.CANCELLED]


_STATUS_TO_COLOR = {
    RequestStatus.PENDING: colorama.Fore.BLUE,
    RequestStatus.RUNNING: colorama.Fore.GREEN,
    RequestStatus.SUCCEEDED: colorama.Fore.GREEN,
    RequestStatus.FAILED: colorama.Fore.RED,
    RequestStatus.CANCELLED: colorama.Fore.WHITE,
}

REQUEST_COLUMNS = [
    'request_id',
    'name',
    'entrypoint',
    'request_body',
    'status',
    'return_value',
    'error',
    'pid',
    'created_at',
    COL_CLUSTER_NAME,
    'schedule_type',
    COL_USER_ID,
    COL_STATUS_MSG,
    COL_SHOULD_RETRY,
    COL_FINISHED_AT,
]


class ScheduleType(enum.Enum):
    """The schedule type for the requests."""
    LONG = 'long'
    # Queue for requests that should be executed quickly for a quick response.
    SHORT = 'short'


@dataclasses.dataclass
class Request:
    """A SkyPilot API request."""

    request_id: str
    name: str
    entrypoint: Callable
    request_body: payloads.RequestBody
    status: RequestStatus
    created_at: float
    user_id: str
    return_value: Any = None
    error: Optional[Dict[str, Any]] = None
    # The pid of the request worker that is(was) running this request.
    pid: Optional[int] = None
    schedule_type: ScheduleType = ScheduleType.LONG
    # Resources the request operates on.
    cluster_name: Optional[str] = None
    # Status message of the request, indicates the reason of current status.
    status_msg: Optional[str] = None
    # Whether the request should be retried.
    should_retry: bool = False
    # When the request finished.
    finished_at: Optional[float] = None

    @property
    def log_path(self) -> pathlib.Path:
        log_path_prefix = pathlib.Path(
            REQUEST_LOG_PATH_PREFIX).expanduser().absolute()
        log_path_prefix.mkdir(parents=True, exist_ok=True)
        log_path = (log_path_prefix / self.request_id).with_suffix('.log')
        return log_path

    def set_error(self, error: BaseException) -> None:
        """Set the error."""
        # TODO(zhwu): pickle.dump does not work well with custom exceptions if
        # it has more than 1 arguments.
        serialized = exceptions.serialize_exception(error)
        self.error = {
            'object': encoders.pickle_and_encode(serialized),
            'type': type(error).__name__,
            'message': str(error),
        }

    def get_error(self) -> Optional[Dict[str, Any]]:
        """Get the error."""
        if self.error is None:
            return None
        unpickled = decoders.decode_and_unpickle(self.error['object'])
        deserialized = exceptions.deserialize_exception(unpickled)
        return {
            'object': deserialized,
            'type': self.error['type'],
            'message': self.error['message'],
        }

    def set_return_value(self, return_value: Any) -> None:
        """Set the return value."""
        self.return_value = encoders.get_encoder(self.name)(return_value)

    def get_return_value(self) -> Any:
        """Get the return value."""
        return decoders.get_decoder(self.name)(self.return_value)

    @classmethod
    def from_row(cls, row: Tuple[Any, ...]) -> 'Request':
        content = dict(zip(REQUEST_COLUMNS, row))
        return cls.decode(payloads.RequestPayload(**content))

    def to_row(self) -> Tuple[Any, ...]:
        payload = self.encode()
        row = []
        for k in REQUEST_COLUMNS:
            row.append(getattr(payload, k))
        return tuple(row)

    def readable_encode(self) -> payloads.RequestPayload:
        """Serialize the SkyPilot API request for display purposes.

        This function should be called on the server side to serialize the
        request body into human readable format, e.g., the entrypoint should
        be a string, and the pid, error, or return value are not needed.

        The returned value will then be displayed on the client side in request
        table.

        We do not use `encode` for display to avoid a large amount of data being
        sent to the client side, especially for the request table could include
        all the requests.
        """
        assert isinstance(self.request_body,
                          payloads.RequestBody), (self.name, self.request_body)
        user = global_user_state.get_user(self.user_id)
        user_name = user.name if user is not None else None
        return payloads.RequestPayload(
            request_id=self.request_id,
            name=self.name,
            entrypoint=self.entrypoint.__name__,
            request_body=self.request_body.model_dump_json(),
            status=self.status.value,
            return_value=json.dumps(None),
            error=json.dumps(None),
            pid=None,
            created_at=self.created_at,
            schedule_type=self.schedule_type.value,
            user_id=self.user_id,
            user_name=user_name,
            cluster_name=self.cluster_name,
            status_msg=self.status_msg,
            should_retry=self.should_retry,
            finished_at=self.finished_at,
        )

    def encode(self) -> payloads.RequestPayload:
        """Serialize the SkyPilot API request."""
        assert isinstance(self.request_body,
                          payloads.RequestBody), (self.name, self.request_body)
        try:
            return payloads.RequestPayload(
                request_id=self.request_id,
                name=self.name,
                entrypoint=encoders.pickle_and_encode(self.entrypoint),
                request_body=encoders.pickle_and_encode(self.request_body),
                status=self.status.value,
                return_value=json.dumps(self.return_value),
                error=json.dumps(self.error),
                pid=self.pid,
                created_at=self.created_at,
                schedule_type=self.schedule_type.value,
                user_id=self.user_id,
                cluster_name=self.cluster_name,
                status_msg=self.status_msg,
                should_retry=self.should_retry,
                finished_at=self.finished_at,
            )
        except (TypeError, ValueError) as e:
            # The error is unexpected, so we don't suppress the stack trace.
            logger.error(
                f'Error encoding: {e}\n'
                f'  {self.request_id}\n'
                f'  {self.name}\n'
                f'  {self.request_body}\n'
                f'  {self.return_value}\n'
                f'  {self.created_at}\n',
                exc_info=e)
            raise

    @classmethod
    def decode(cls, payload: payloads.RequestPayload) -> 'Request':
        """Deserialize the SkyPilot API request."""
        try:
            return cls(
                request_id=payload.request_id,
                name=payload.name,
                entrypoint=decoders.decode_and_unpickle(payload.entrypoint),
                request_body=decoders.decode_and_unpickle(payload.request_body),
                status=RequestStatus(payload.status),
                return_value=json.loads(payload.return_value),
                error=json.loads(payload.error),
                pid=payload.pid,
                created_at=payload.created_at,
                schedule_type=ScheduleType(payload.schedule_type),
                user_id=payload.user_id,
                cluster_name=payload.cluster_name,
                status_msg=payload.status_msg,
                should_retry=payload.should_retry,
                finished_at=payload.finished_at,
            )
        except (TypeError, ValueError) as e:
            logger.error(
                f'Error decoding: {e}\n'
                f'  {payload.request_id}\n'
                f'  {payload.name}\n'
                f'  {payload.entrypoint}\n'
                f'  {payload.request_body}\n'
                f'  {payload.created_at}\n',
                exc_info=e)
            # The error is unexpected, so we don't suppress the stack trace.
            raise


def kill_cluster_requests(cluster_name: str, exclude_request_name: str):
    """Kill all pending and running requests for a cluster.

    Args:
        cluster_name: the name of the cluster.
        exclude_request_names: exclude requests with these names. This is to
            prevent killing the caller request.
    """
    request_ids = [
        request_task.request_id for request_task in get_request_tasks(
            cluster_names=[cluster_name],
            status=[RequestStatus.PENDING, RequestStatus.RUNNING],
            exclude_request_names=[exclude_request_name])
    ]
    kill_requests(request_ids)


def kill_requests(request_ids: Optional[List[str]] = None,
                  user_id: Optional[str] = None) -> List[str]:
    """Kill a SkyPilot API request and set its status to cancelled.

    Args:
        request_ids: The request IDs to kill. If None, all requests for the
            user are killed.
        user_id: The user ID to kill requests for. If None, all users are
            killed.

    Returns:
        A list of request IDs that were cancelled.
    """
    if request_ids is None:
        request_ids = [
            request_task.request_id for request_task in get_request_tasks(
                user_id=user_id,
                status=[RequestStatus.RUNNING, RequestStatus.PENDING],
                # Avoid cancelling the cancel request itself.
                exclude_request_names=['sky.api_cancel'])
        ]
    cancelled_request_ids = []
    for request_id in request_ids:
        with update_request(request_id) as request_record:
            if request_record is None:
                logger.debug(f'No request ID {request_id}')
                continue
            # Skip internal requests. The internal requests are scheduled with
            # request_id in range(len(INTERNAL_REQUEST_EVENTS)).
            if request_record.request_id in set(
                    event.id for event in daemons.INTERNAL_REQUEST_DAEMONS):
                continue
            if request_record.status > RequestStatus.RUNNING:
                logger.debug(f'Request {request_id} already finished')
                continue
            if request_record.pid is not None:
                logger.debug(f'Killing request process {request_record.pid}')
                # Use SIGTERM instead of SIGKILL:
                # - The executor can handle SIGTERM gracefully
                # - After SIGTERM, the executor can reuse the request process
                #   for other requests, avoiding the overhead of forking a new
                #   process for each request.
                os.kill(request_record.pid, signal.SIGTERM)
            request_record.status = RequestStatus.CANCELLED
            request_record.finished_at = time.time()
            cancelled_request_ids.append(request_id)
    return cancelled_request_ids


def create_table(cursor, conn):
    # Enable WAL mode to avoid locking issues.
    # See: issue #1441 and PR #1509
    # https://github.com/microsoft/WSL/issues/2395
    # TODO(romilb): We do not enable WAL for WSL because of known issue in WSL.
    #  This may cause the database locked problem from WSL issue #1441.
    if not common_utils.is_wsl():
        try:
            cursor.execute('PRAGMA journal_mode=WAL')
        except sqlite3.OperationalError as e:
            if 'database is locked' not in str(e):
                raise
            # If the database is locked, it is OK to continue, as the WAL mode
            # is not critical and is likely to be enabled by other processes.

    # Table for Requests
    cursor.execute(f"""\
        CREATE TABLE IF NOT EXISTS {REQUEST_TABLE} (
        request_id TEXT PRIMARY KEY,
        name TEXT,
        entrypoint TEXT,
        request_body TEXT,
        status TEXT,
        created_at REAL,
        return_value TEXT,
        error BLOB,
        pid INTEGER,
        {COL_CLUSTER_NAME} TEXT,
        schedule_type TEXT,
        {COL_USER_ID} TEXT,
        {COL_STATUS_MSG} TEXT,
        {COL_SHOULD_RETRY} INTEGER,
        {COL_FINISHED_AT} REAL
        )""")

    db_utils.add_column_to_table(cursor, conn, REQUEST_TABLE, COL_STATUS_MSG,
                                 'TEXT')
    db_utils.add_column_to_table(cursor, conn, REQUEST_TABLE, COL_SHOULD_RETRY,
                                 'INTEGER')
    db_utils.add_column_to_table(cursor, conn, REQUEST_TABLE, COL_FINISHED_AT,
                                 'REAL')


_DB = None
_init_db_lock = threading.Lock()


def init_db(func):
    """Initialize the database."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        global _DB
        if _DB is not None:
            return func(*args, **kwargs)
        with _init_db_lock:
            if _DB is None:
                db_path = os.path.expanduser(
                    server_constants.API_SERVER_REQUEST_DB_PATH)
                pathlib.Path(db_path).parents[0].mkdir(parents=True,
                                                       exist_ok=True)
                _DB = db_utils.SQLiteConn(db_path, create_table)
        return func(*args, **kwargs)

    return wrapper


def reset_db_and_logs():
    """Create the database."""
    server_common.clear_local_api_server_database()
    shutil.rmtree(pathlib.Path(REQUEST_LOG_PATH_PREFIX).expanduser(),
                  ignore_errors=True)
    shutil.rmtree(server_common.API_SERVER_CLIENT_DIR.expanduser(),
                  ignore_errors=True)


def request_lock_path(request_id: str) -> str:
    lock_path = os.path.expanduser(REQUEST_LOG_PATH_PREFIX)
    os.makedirs(lock_path, exist_ok=True)
    return os.path.join(lock_path, f'.{request_id}.lock')


@contextlib.contextmanager
@init_db
def update_request(request_id: str) -> Generator[Optional[Request], None, None]:
    """Get a SkyPilot API request."""
    request = _get_request_no_lock(request_id)
    yield request
    if request is not None:
        _add_or_update_request_no_lock(request)


def _get_request_no_lock(request_id: str) -> Optional[Request]:
    """Get a SkyPilot API request."""
    assert _DB is not None
    columns_str = ', '.join(REQUEST_COLUMNS)
    with _DB.conn:
        cursor = _DB.conn.cursor()
        cursor.execute(
            f'SELECT {columns_str} FROM {REQUEST_TABLE} '
            'WHERE request_id LIKE ?', (request_id + '%',))
        row = cursor.fetchone()
        if row is None:
            return None
    return Request.from_row(row)


@init_db
def get_latest_request_id() -> Optional[str]:
    """Get the latest request ID."""
    assert _DB is not None
    with _DB.conn:
        cursor = _DB.conn.cursor()
        cursor.execute(f'SELECT request_id FROM {REQUEST_TABLE} '
                       'ORDER BY created_at DESC LIMIT 1')
        row = cursor.fetchone()
        return row[0] if row else None


@init_db
def get_request(request_id: str) -> Optional[Request]:
    """Get a SkyPilot API request."""
    with filelock.FileLock(request_lock_path(request_id)):
        return _get_request_no_lock(request_id)


@init_db
def create_if_not_exists(request: Request) -> bool:
    """Create a SkyPilot API request if it does not exist."""
    with filelock.FileLock(request_lock_path(request.request_id)):
        if _get_request_no_lock(request.request_id) is not None:
            return False
        _add_or_update_request_no_lock(request)
        return True


@init_db
def get_request_tasks(
    status: Optional[List[RequestStatus]] = None,
    cluster_names: Optional[List[str]] = None,
    user_id: Optional[str] = None,
    exclude_request_names: Optional[List[str]] = None,
    include_request_names: Optional[List[str]] = None,
    finished_before: Optional[float] = None,
) -> List[Request]:
    """Get a list of requests that match the given filters.

    Args:
        status: a list of statuses of the requests to filter on.
        cluster_names: a list of cluster names to filter requests on.
        exclude_request_names: a list of request names to exclude from results.
            Mutually exclusive with include_request_names.
        user_id: the user ID to filter requests on.
            If None, all users are included.
        include_request_names: a list of request names to filter on.
            Mutually exclusive with exclude_request_names.
        finished_before: if provided, only include requests finished before this
            timestamp.

    Raises:
        ValueError: If both exclude_request_names and include_request_names are
            provided.
    """
    if exclude_request_names is not None and include_request_names is not None:
        raise ValueError(
            'Only one of exclude_request_names or include_request_names can be '
            'provided, not both.')

    filters = []
    filter_params: List[Any] = []
    if status is not None:
        status_list_str = ','.join(repr(status.value) for status in status)
        filters.append(f'status IN ({status_list_str})')
    if exclude_request_names is not None:
        exclude_request_names_str = ','.join(
            repr(name) for name in exclude_request_names)
        filters.append(f'name NOT IN ({exclude_request_names_str})')
    if cluster_names is not None:
        cluster_names_str = ','.join(repr(name) for name in cluster_names)
        filters.append(f'{COL_CLUSTER_NAME} IN ({cluster_names_str})')
    if user_id is not None:
        filters.append(f'{COL_USER_ID} = ?')
        filter_params.append(user_id)
    if include_request_names is not None:
        request_names_str = ','.join(
            repr(name) for name in include_request_names)
        filters.append(f'name IN ({request_names_str})')
    if finished_before is not None:
        filters.append('finished_at < ?')
        filter_params.append(finished_before)
    assert _DB is not None
    with _DB.conn:
        cursor = _DB.conn.cursor()
        filter_str = ' AND '.join(filters)
        if filter_str:
            filter_str = f' WHERE {filter_str}'
        columns_str = ', '.join(REQUEST_COLUMNS)
        cursor.execute(
            f'SELECT {columns_str} FROM {REQUEST_TABLE}{filter_str} '
            'ORDER BY created_at DESC', filter_params)
        rows = cursor.fetchall()
        if rows is None:
            return []
    requests = []
    for row in rows:
        request = Request.from_row(row)
        requests.append(request)
    return requests


def _add_or_update_request_no_lock(request: Request):
    """Add or update a REST request into the database."""
    row = request.to_row()
    key_str = ', '.join(REQUEST_COLUMNS)
    fill_str = ', '.join(['?'] * len(row))
    assert _DB is not None
    with _DB.conn:
        cursor = _DB.conn.cursor()
        cursor.execute(
            f'INSERT OR REPLACE INTO {REQUEST_TABLE} ({key_str}) '
            f'VALUES ({fill_str})', row)


def set_request_failed(request_id: str, e: BaseException) -> None:
    """Set a request to failed and populate the error message."""
    with ux_utils.enable_traceback():
        stacktrace = traceback.format_exc()
    setattr(e, 'stacktrace', stacktrace)
    with update_request(request_id) as request_task:
        assert request_task is not None, request_id
        request_task.status = RequestStatus.FAILED
        request_task.finished_at = time.time()
        request_task.set_error(e)


def set_request_succeeded(request_id: str, result: Optional[Any]) -> None:
    """Set a request to succeeded and populate the result."""
    with update_request(request_id) as request_task:
        assert request_task is not None, request_id
        request_task.status = RequestStatus.SUCCEEDED
        request_task.finished_at = time.time()
        if result is not None:
            request_task.set_return_value(result)


def set_request_cancelled(request_id: str) -> None:
    """Set a request to cancelled."""
    with update_request(request_id) as request_task:
        assert request_task is not None, request_id
        request_task.finished_at = time.time()
        request_task.status = RequestStatus.CANCELLED


@init_db
def _delete_requests(requests: List[Request]):
    """Clean up requests by their IDs."""
    id_list_str = ','.join(repr(req.request_id) for req in requests)
    assert _DB is not None
    with _DB.conn:
        cursor = _DB.conn.cursor()
        cursor.execute(
            f'DELETE FROM {REQUEST_TABLE} WHERE request_id IN ({id_list_str})')


def clean_finished_requests_with_retention(retention_seconds: int):
    """Clean up finished requests older than the retention period.

    This function removes old finished requests (SUCCEEDED, FAILED, CANCELLED)
    from the database and cleans up their associated log files.

    Args:
        retention_seconds: Requests older than this many seconds will be
            deleted.
    """
    reqs = get_request_tasks(status=RequestStatus.finished_status(),
                             finished_before=time.time() - retention_seconds)

    subprocess_utils.run_in_parallel(
        func=lambda req: req.log_path.unlink(missing_ok=True),
        args=reqs,
        num_threads=len(reqs))

    _delete_requests(reqs)

    # To avoid leakage of the log file, logs must be deleted before the
    # request task in the database.
    logger.info(f'Cleaned up {len(reqs)} finished requests '
                f'older than {retention_seconds} seconds')


async def requests_gc_daemon():
    """Garbage collect finished requests periodically."""
    while True:
        logger.info('Running requests GC daemon...')
        # Use the latest config.
        skypilot_config.reload_config()
        retention_seconds = skypilot_config.get_nested(
            ('api_server', 'requests_retention_hours'),
            DEFAULT_REQUESTS_RETENTION_HOURS) * 3600
        try:
            # Negative value disables the requests GC
            if retention_seconds >= 0:
                clean_finished_requests_with_retention(retention_seconds)
        except asyncio.CancelledError:
            logger.info('Requests GC daemon cancelled')
            break
        except Exception as e:  # pylint: disable=broad-except
            logger.error(f'Error running requests GC daemon: {e}')
        # Run the daemon at most once every hour to avoid too frequent
        # cleanup.
        await asyncio.sleep(max(retention_seconds, 3600))



"""Strategies to handle launching/recovery/termination of managed job clusters.

In the YAML file, the user can specify the strategy to use for managed jobs.

resources:
    job_recovery: EAGER_NEXT_REGION
"""
import time
import traceback
import typing
from typing import Optional

from sky import backends
from sky import dag as dag_lib
from sky import exceptions
from sky import execution
from sky import global_user_state
from sky import sky_logging
from sky.backends import backend_utils
from sky.jobs import scheduler
from sky.jobs import state
from sky.jobs import utils as managed_job_utils
from sky.serve import serve_utils
from sky.skylet import job_lib
from sky.usage import usage_lib
from sky.utils import common_utils
from sky.utils import registry
from sky.utils import status_lib
from sky.utils import ux_utils

if typing.TYPE_CHECKING:
    from sky import resources
    from sky import task as task_lib

logger = sky_logging.init_logger(__name__)

# Waiting time for job from INIT/PENDING to RUNNING
# 10 * JOB_STARTED_STATUS_CHECK_GAP_SECONDS = 10 * 5 = 50 seconds
MAX_JOB_CHECKING_RETRY = 10

# Minutes to job cluster autodown. This should be significantly larger than
# managed_job_utils.JOB_STATUS_CHECK_GAP_SECONDS, to avoid tearing down the
# cluster before its status can be updated by the job controller.
_AUTODOWN_MINUTES = 5


class StrategyExecutor:
    """Handle the launching, recovery and termination of managed job clusters"""

    RETRY_INIT_GAP_SECONDS = 60

    def __init__(self, cluster_name: Optional[str], backend: 'backends.Backend',
                 task: 'task_lib.Task', max_restarts_on_errors: int,
                 job_id: int, task_id: int, pool: Optional[str]) -> None:
        """Initialize the strategy executor.

        Args:
            cluster_name: The name of the cluster.
            backend: The backend to use. Only CloudVMRayBackend is supported.
            task: The task to execute.
        """
        assert isinstance(backend, backends.CloudVmRayBackend), (
            'Only CloudVMRayBackend is supported.')
        self.dag = dag_lib.Dag()
        self.dag.add(task)
        # For jobs submitted to a pool, the cluster name might change after each
        # recovery. Initially this is set to an empty string to indicate that no
        # cluster is assigned yet, and in `_launch`, it will be set to one of
        # the cluster names in the pool.
        self.cluster_name = cluster_name
        self.backend = backend
        self.max_restarts_on_errors = max_restarts_on_errors
        self.job_id = job_id
        self.task_id = task_id
        self.pool = pool
        self.restart_cnt_on_failure = 0
        self.job_id_on_pool_cluster: Optional[int] = None

    @classmethod
    def make(cls, cluster_name: Optional[str], backend: 'backends.Backend',
             task: 'task_lib.Task', job_id: int, task_id: int,
             pool: Optional[str]) -> 'StrategyExecutor':
        """Create a strategy from a task."""

        resource_list = list(task.resources)
        job_recovery = resource_list[0].job_recovery
        for resource in resource_list:
            if resource.job_recovery != job_recovery:
                raise ValueError(
                    'The job recovery strategy should be the same for all '
                    'resources.')
        # Remove the job_recovery field from the resources, as the strategy
        # will be handled by the strategy class.
        new_resources_list = [r.copy(job_recovery=None) for r in resource_list]
        # set the new_task_resources to be the same type (list or set) as the
        # original task.resources
        task.set_resources(type(task.resources)(new_resources_list))
        if isinstance(job_recovery, dict):
            name = job_recovery.pop(
                'strategy', registry.JOBS_RECOVERY_STRATEGY_REGISTRY.default)
            assert name is None or isinstance(name, str), (
                name, 'The job recovery strategy name must be a string or None')
            job_recovery_name: Optional[str] = name
            max_restarts_on_errors = job_recovery.pop('max_restarts_on_errors',
                                                      0)
        else:
            job_recovery_name = job_recovery
            max_restarts_on_errors = 0
        job_recovery_strategy = (registry.JOBS_RECOVERY_STRATEGY_REGISTRY.
                                 from_str(job_recovery_name))
        assert job_recovery_strategy is not None, job_recovery_name
        return job_recovery_strategy(cluster_name, backend, task,
                                     max_restarts_on_errors, job_id, task_id,
                                     pool)

    def launch(self) -> float:
        """Launch the cluster for the first time.

        It can fail if resource is not available. Need to check the cluster
        status, after calling.

        Returns: The job's submit timestamp, on success (otherwise, an
            exception is raised).

        Raises: Please refer to the docstring of self._launch().
        """

        job_submit_at = self._launch(max_retry=None)
        assert job_submit_at is not None
        return job_submit_at

    def recover(self) -> float:
        """Relaunch the cluster after failure and wait until job starts.

        When recover() is called the cluster should be in STOPPED status (i.e.
        partially down).

        Returns: The timestamp job started.
        """
        raise NotImplementedError

    def _try_cancel_jobs(self):
        from sky import core  # pylint: disable=import-outside-toplevel

        if self.cluster_name is None:
            return
        handle = global_user_state.get_handle_from_cluster_name(
            self.cluster_name)
        if handle is None or self.pool is not None:
            return
        try:
            usage_lib.messages.usage.set_internal()
            # Note that `sky.cancel()` may not go through for a variety of
            # reasons:
            # (1) head node is preempted; or
            # (2) somehow user programs escape the cancel codepath's kill.
            # The latter is silent and is a TODO.
            #
            # For the former, an exception will be thrown, in which case we
            # fallback to terminate_cluster() in the except block below. This
            # is because in the event of recovery on the same set of remaining
            # worker nodes, we don't want to leave some old job processes
            # running.
            # TODO(zhwu): This is non-ideal and we should figure out another way
            # to reliably cancel those processes and not have to down the
            # remaining nodes first.
            #
            # In the case where the worker node is preempted, the `sky.cancel()`
            # should be functional with the `_try_cancel_if_cluster_is_init`
            # flag, i.e. it sends the cancel signal to the head node, which will
            # then kill the user process on remaining worker nodes.
            # Only cancel the corresponding job for worker pool.
            if self.pool is None:
                kwargs = dict(all=True)
            else:
                kwargs = dict(job_ids=[self.job_id_on_pool_cluster])
            core.cancel(cluster_name=self.cluster_name,
                        **kwargs,
                        _try_cancel_if_cluster_is_init=True)
        except Exception as e:  # pylint: disable=broad-except
            logger.info('Failed to cancel the job on the cluster. The cluster '
                        'might be already down or the head node is preempted.'
                        '\n  Detailed exception: '
                        f'{common_utils.format_exception(e)}\n'
                        'Terminating the cluster explicitly to ensure no '
                        'remaining job process interferes with recovery.')
            self._cleanup_cluster()

    def _wait_until_job_starts_on_cluster(self) -> Optional[float]:
        """Wait for MAX_JOB_CHECKING_RETRY times until job starts on the cluster

        Returns:
            The timestamp of when the job is submitted, or None if failed to
            submit.
        """
        assert self.cluster_name is not None
        status = None
        job_checking_retry_cnt = 0
        while job_checking_retry_cnt < MAX_JOB_CHECKING_RETRY:
            # Avoid the infinite loop, if any bug happens.
            job_checking_retry_cnt += 1
            try:
                cluster_status, _ = (
                    backend_utils.refresh_cluster_status_handle(
                        self.cluster_name,
                        force_refresh_statuses=set(status_lib.ClusterStatus)))
            except Exception as e:  # pylint: disable=broad-except
                # If any unexpected error happens, retry the job checking
                # loop.
                # TODO(zhwu): log the unexpected error to usage collection
                # for future debugging.
                logger.info(f'Unexpected exception: {e}\nFailed to get the '
                            'refresh the cluster status. Retrying.')
                continue
            if cluster_status != status_lib.ClusterStatus.UP:
                # The cluster can be preempted before the job is
                # launched.
                # Break to let the retry launch kick in.
                logger.info('The cluster is preempted before the job '
                            'is submitted.')
                # TODO(zhwu): we should recover the preemption with the
                # recovery strategy instead of the current while loop.
                break

            try:
                status = managed_job_utils.get_job_status(
                    self.backend,
                    self.cluster_name,
                    job_id=self.job_id_on_pool_cluster)
            except Exception as e:  # pylint: disable=broad-except
                # If any unexpected error happens, retry the job checking
                # loop.
                # Note: the CommandError is already handled in the
                # get_job_status, so it should not happen here.
                # TODO(zhwu): log the unexpected error to usage collection
                # for future debugging.
                logger.info(f'Unexpected exception: {e}\nFailed to get the '
                            'job status. Retrying.')
                continue

            # Check the job status until it is not in initialized status
            if status is not None and status > job_lib.JobStatus.INIT:
                try:
                    job_submitted_at = managed_job_utils.get_job_timestamp(
                        self.backend,
                        self.cluster_name,
                        self.job_id_on_pool_cluster,
                        get_end_time=False)
                    return job_submitted_at
                except Exception as e:  # pylint: disable=broad-except
                    # If we failed to get the job timestamp, we will retry
                    # job checking loop.
                    logger.info(f'Unexpected Exception: {e}\nFailed to get '
                                'the job start timestamp. Retrying.')
                    continue
            # Wait for the job to be started
            time.sleep(managed_job_utils.JOB_STARTED_STATUS_CHECK_GAP_SECONDS)
        return None

    def _cleanup_cluster(self) -> None:
        if self.cluster_name is None:
            return
        if self.pool is None:
            managed_job_utils.terminate_cluster(self.cluster_name)

    def _launch(self,
                max_retry: Optional[int] = 3,
                raise_on_failure: bool = True,
                recovery: bool = False) -> Optional[float]:
        """Implementation of launch().

        The function will wait until the job starts running, but will leave the
        handling for the preemption to the caller.

        Args:
            max_retry: The maximum number of retries. If None, retry forever.
            raise_on_failure: Whether to raise an exception if the launch fails.

        Returns:
            The job's submit timestamp, or None if failed to submit the job
            (either provisioning fails or any error happens in job submission)
            and raise_on_failure is False.

        Raises:
            non-exhaustive list of exceptions:
            exceptions.ProvisionPrechecksError: This will be raised when the
                underlying `sky.launch` fails due to precheck errors only.
                I.e., none of the failover exceptions, if
                any, is due to resources unavailability. This exception
                includes the following cases:
                1. The optimizer cannot find a feasible solution.
                2. Precheck errors: invalid cluster name, failure in getting
                cloud user identity, or unsupported feature.
            exceptions.ManagedJobReachedMaxRetriesError: This will be raised
                when all prechecks passed but the maximum number of retries is
                reached for `sky.launch`. The failure of `sky.launch` can be
                due to:
                1. Any of the underlying failover exceptions is due to resources
                unavailability.
                2. The cluster is preempted before the job is submitted.
                3. Any unexpected error happens during the `sky.launch`.
        Other exceptions may be raised depending on the backend.
        """
        # TODO(zhwu): handle the failure during `preparing sky runtime`.
        retry_cnt = 0
        backoff = common_utils.Backoff(self.RETRY_INIT_GAP_SECONDS)
        while True:
            retry_cnt += 1
            try:
                with scheduler.scheduled_launch(self.job_id):
                    # The job state may have been PENDING during backoff -
                    # update to STARTING or RECOVERING.
                    # On the first attempt (when retry_cnt is 1), we should
                    # already be in STARTING or RECOVERING.
                    if retry_cnt > 1:
                        state.set_restarting(self.job_id, self.task_id,
                                             recovery)
                    try:
                        usage_lib.messages.usage.set_internal()
                        if self.pool is None:
                            assert self.cluster_name is not None
                            # Detach setup, so that the setup failure can be
                            # detected by the controller process (job_status ->
                            # FAILED_SETUP).
                            execution.launch(
                                self.dag,
                                cluster_name=self.cluster_name,
                                # We expect to tear down the cluster as soon as
                                # the job is finished. However, in case the
                                # controller dies, set autodown to try and avoid
                                # a resource leak.
                                idle_minutes_to_autostop=_AUTODOWN_MINUTES,
                                down=True,
                                _is_launched_by_jobs_controller=True)
                        else:
                            self.cluster_name = (
                                serve_utils.get_next_cluster_name(
                                    self.pool, self.job_id))
                            if self.cluster_name is None:
                                raise exceptions.NoClusterLaunchedError(
                                    'No cluster name found in the pool.')
                            job_id_on_pool_cluster, _ = execution.exec(
                                self.dag, cluster_name=self.cluster_name)
                            assert job_id_on_pool_cluster is not None, (
                                self.cluster_name, self.job_id)
                            self.job_id_on_pool_cluster = job_id_on_pool_cluster
                            state.set_job_id_on_pool_cluster(
                                self.job_id, job_id_on_pool_cluster)
                        logger.info('Managed job cluster launched.')
                    except (exceptions.InvalidClusterNameError,
                            exceptions.NoCloudAccessError,
                            exceptions.ResourcesMismatchError) as e:
                        logger.error('Failure happened before provisioning. '
                                     f'{common_utils.format_exception(e)}')
                        if raise_on_failure:
                            raise exceptions.ProvisionPrechecksError(
                                reasons=[e])
                        return None
                    except exceptions.ResourcesUnavailableError as e:
                        # This is raised when the launch fails due to prechecks
                        # or after failing over through all the candidates.
                        # Please refer to the docstring of `sky.launch` for more
                        # details of how the exception will be structured.
                        if not any(
                                isinstance(err,
                                           exceptions.ResourcesUnavailableError)
                                for err in e.failover_history):
                            # _launch() (this function) should fail/exit
                            # directly, if none of the failover reasons were
                            # because of resource unavailability or no failover
                            # was attempted (the optimizer cannot find feasible
                            # resources for requested resources), i.e.,
                            # e.failover_history is empty. Failing directly
                            # avoids the infinite loop of retrying the launch
                            # when, e.g., an invalid cluster name is used and
                            # --retry-until-up is specified.
                            reasons = (e.failover_history
                                       if e.failover_history else [e])
                            reasons_str = '; '.join(
                                common_utils.format_exception(err)
                                for err in reasons)
                            logger.error(
                                'Failure happened before provisioning. '
                                f'Failover reasons: {reasons_str}')
                            if raise_on_failure:
                                raise exceptions.ProvisionPrechecksError(
                                    reasons)
                            return None
                        logger.info('Failed to launch a cluster with error: '
                                    f'{common_utils.format_exception(e)})')
                    except Exception as e:  # pylint: disable=broad-except
                        # If the launch fails, it will be recovered by the
                        # following code.
                        logger.info('Failed to launch a cluster with error: '
                                    f'{common_utils.format_exception(e)})')
                        with ux_utils.enable_traceback():
                            logger.info(
                                f'  Traceback: {traceback.format_exc()}')
                    else:  # No exception, the launch succeeds.
                        # At this point, a sky.launch() has succeeded. Cluster
                        # may be UP (no preemption since) or DOWN (newly
                        # preempted).
                        job_submitted_at = (
                            self._wait_until_job_starts_on_cluster())
                        if job_submitted_at is not None:
                            return job_submitted_at
                        # The job fails to start on the cluster, retry the
                        # launch.
                        # TODO(zhwu): log the unexpected error to usage
                        # collection for future debugging.
                        logger.info(
                            'Failed to successfully submit the job to the '
                            'launched cluster, due to unexpected submission '
                            'errors or the cluster being preempted during '
                            'job submission.')

                    # If we get here, the launch did not succeed. Tear down the
                    # cluster and retry.
                    self._cleanup_cluster()
                    if max_retry is not None and retry_cnt >= max_retry:
                        # Retry forever if max_retry is None.
                        if raise_on_failure:
                            with ux_utils.print_exception_no_traceback():
                                raise (
                                    exceptions.ManagedJobReachedMaxRetriesError(
                                        'Resources unavailable: failed to '
                                        f'launch clusters after {max_retry} '
                                        'retries.'))
                        else:
                            return None

                    # Raise NoClusterLaunchedError to indicate that the job is
                    # in retry backoff. This will trigger special handling in
                    # scheduler.schedule_launched().
                    # We will exit the scheduled_launch context so that the
                    # schedule state is ALIVE_BACKOFF during the backoff. This
                    # allows other jobs to launch.
                    raise exceptions.NoClusterLaunchedError()

            except exceptions.NoClusterLaunchedError:
                # Update the status to PENDING during backoff.
                state.set_backoff_pending(self.job_id, self.task_id)
                # Calculate the backoff time and sleep.
                # We retry immediately for worker pool, since no sky.launch()
                # is called and the overhead is minimal.
                gap_seconds = (backoff.current_backoff()
                               if self.pool is None else 1)
                logger.info('Retrying to launch the cluster in '
                            f'{gap_seconds:.1f} seconds.')
                time.sleep(gap_seconds)
                continue
            else:
                # The inner loop should either return or throw
                # NoClusterLaunchedError.
                assert False, 'Unreachable'

    def should_restart_on_failure(self) -> bool:
        """Increments counter & checks if job should be restarted on a failure.

        Returns:
            True if the job should be restarted, otherwise False.
        """
        self.restart_cnt_on_failure += 1
        if self.restart_cnt_on_failure > self.max_restarts_on_errors:
            return False
        return True


@registry.JOBS_RECOVERY_STRATEGY_REGISTRY.type_register(name='FAILOVER',
                                                        default=False)
class FailoverStrategyExecutor(StrategyExecutor):
    """Failover strategy: wait in same region and failover after timeout."""

    _MAX_RETRY_CNT = 240  # Retry for 4 hours.

    def __init__(self, cluster_name: Optional[str], backend: 'backends.Backend',
                 task: 'task_lib.Task', max_restarts_on_errors: int,
                 job_id: int, task_id: int, pool: Optional[str]) -> None:
        super().__init__(cluster_name, backend, task, max_restarts_on_errors,
                         job_id, task_id, pool)
        # Note down the cloud/region of the launched cluster, so that we can
        # first retry in the same cloud/region. (Inside recover() we may not
        # rely on cluster handle, as it can be None if the cluster is
        # preempted.)
        self._launched_resources: Optional['resources.Resources'] = None

    def _launch(self,
                max_retry: Optional[int] = 3,
                raise_on_failure: bool = True,
                recovery: bool = False) -> Optional[float]:
        job_submitted_at = super()._launch(max_retry, raise_on_failure,
                                           recovery)
        if job_submitted_at is not None and self.cluster_name is not None:
            # Only record the cloud/region if the launch is successful.
            handle = global_user_state.get_handle_from_cluster_name(
                self.cluster_name)
            assert isinstance(handle, backends.CloudVmRayResourceHandle), (
                'Cluster should be launched.', handle)
            launched_resources = handle.launched_resources
            self._launched_resources = launched_resources
        else:
            self._launched_resources = None
        return job_submitted_at

    def recover(self) -> float:
        # 1. Cancel the jobs and launch the cluster with the STOPPED status,
        #    so that it will try on the current region first until timeout.
        # 2. Tear down the cluster, if the step 1 failed to launch the cluster.
        # 3. Launch the cluster with no cloud/region constraint or respect the
        #    original user specification.

        # Step 1
        self._try_cancel_jobs()

        while True:
            # Add region constraint to the task, to retry on the same region
            # first (if valid).
            if self._launched_resources is not None:
                task = self.dag.tasks[0]
                original_resources = task.resources
                launched_cloud = self._launched_resources.cloud
                launched_region = self._launched_resources.region
                new_resources = self._launched_resources.copy(
                    cloud=launched_cloud, region=launched_region, zone=None)
                task.set_resources({new_resources})
                # Not using self.launch to avoid the retry until up logic.
                job_submitted_at = self._launch(raise_on_failure=False,
                                                recovery=True)
                # Restore the original dag, i.e. reset the region constraint.
                task.set_resources(original_resources)
                if job_submitted_at is not None:
                    return job_submitted_at

            # Step 2
            logger.debug('Terminating unhealthy cluster and reset cloud '
                         'region.')
            self._cleanup_cluster()

            # Step 3
            logger.debug('Relaunch the cluster  without constraining to prior '
                         'cloud/region.')
            # Not using self.launch to avoid the retry until up logic.
            job_submitted_at = self._launch(max_retry=self._MAX_RETRY_CNT,
                                            raise_on_failure=False,
                                            recovery=True)
            if job_submitted_at is None:
                # Failed to launch the cluster.
                gap_seconds = self.RETRY_INIT_GAP_SECONDS
                logger.info('Retrying to recover the cluster in '
                            f'{gap_seconds:.1f} seconds.')
                time.sleep(gap_seconds)
                continue

            return job_submitted_at


@registry.JOBS_RECOVERY_STRATEGY_REGISTRY.type_register(
    name='EAGER_NEXT_REGION', default=True)
class EagerFailoverStrategyExecutor(FailoverStrategyExecutor):
    """Eager failover strategy.

    This strategy is an extension of the FAILOVER strategy. Instead of waiting
    in the same region when the preemption happens, it immediately terminates
    the cluster and relaunches it in a different region. This is based on the
    observation that the preemption is likely to happen again shortly in the
    same region, so trying other regions first is more likely to get a longer
    running cluster.

    Example: Assume the user has access to 3 regions, R1, R2, R3, in that price
    order. Then the following are some possible event sequences:

        R1Z1 (preempted) -> R2 (success)

        R1Z1 (preempted) -> R2 (failed to launch) -> R3 (success)

        R1Z1 (preempted) -> R2 (failed to launch) -> R3 (failed to launch)
                                                  -> R1Z2 (success)

        R1Z1 (preempted) -> R2 (failed to launch) -> R3 (failed to launch)
                                                  -> R1Z1 (success)
    """

    def recover(self) -> float:
        # 1. Terminate the current cluster
        # 2. Launch again by explicitly blocking the previously launched region
        # (this will failover through the entire search space except the
        # previously launched region)
        # 3. (If step 2 failed) Retry forever: Launch again with no blocked
        # locations (this will failover through the entire search space)
        #
        # The entire search space is defined by the original task request,
        # task.resources.

        # Step 1
        logger.debug('Terminating unhealthy cluster and reset cloud region.')
        self._cleanup_cluster()

        # Step 2
        logger.debug('Relaunch the cluster skipping the previously launched '
                     'cloud/region.')
        if self._launched_resources is not None:
            task = self.dag.tasks[0]
            requested_resources = self._launched_resources
            if (requested_resources.region is None and
                    requested_resources.zone is None):
                # Optimization: We only block the previously launched region,
                # if the requested resources does not specify a region or zone,
                # because, otherwise, we will spend unnecessary time for
                # skipping the only specified region/zone.
                launched_cloud = self._launched_resources.cloud
                launched_region = self._launched_resources.region
                task.blocked_resources = {
                    requested_resources.copy(cloud=launched_cloud,
                                             region=launched_region)
                }
                # Not using self.launch to avoid the retry until up logic.
                job_submitted_at = self._launch(raise_on_failure=False,
                                                recovery=True)
                task.blocked_resources = None
                if job_submitted_at is not None:
                    return job_submitted_at

        while True:
            # Step 3
            logger.debug('Relaunch the cluster without constraining to prior '
                         'cloud/region.')
            # Not using self.launch to avoid the retry until up logic.
            job_submitted_at = self._launch(max_retry=self._MAX_RETRY_CNT,
                                            raise_on_failure=False,
                                            recovery=True)
            if job_submitted_at is None:
                # Failed to launch the cluster.
                gap_seconds = self.RETRY_INIT_GAP_SECONDS
                logger.info('Retrying to recover the cluster in '
                            f'{gap_seconds:.1f} seconds.')
                time.sleep(gap_seconds)
                continue

            return job_submitted_at
